# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    A test class to check that the features related to noisy simulation are working as expected.
"""

import unittest

from openfermion.ops import QubitOperator

from tangelo.linq import Gate, Circuit, Simulator, backend_info
from tangelo.linq.noisy_simulation import NoiseModel, get_qiskit_noise_dict
from tangelo.helpers.utils import default_simulator, installed_backends

# Noisy simulation: circuits, noise models, references
cn1 = Circuit([Gate('X', target=0)])
cn2 = Circuit([Gate('CNOT', target=1, control=0)])

nmp, nmd, nmc = NoiseModel(), NoiseModel(), NoiseModel()
# nmp: pauli noise with equal probabilities, on X and CNOT gates
nmp.add_quantum_error("X", 'pauli', [1 / 3] * 3)
nmp.add_quantum_error("CNOT", 'pauli', [1 / 3] * 3)
# nmd: depol noise with prob 1. on X and CNOT gates
nmd.add_quantum_error("X", 'depol', 1.)
nmd.add_quantum_error("CNOT", 'depol', 1.)
# nmc: cumulates 2 Pauli noises (here, is equivalent to no noise, as it applies Y twice when X is ran)
nmc.add_quantum_error("X", 'pauli', [0., 1., 0.])
nmc.add_quantum_error("X", 'depol', 4/3)

ref_pauli1 = {'1': 1 / 3, '0': 2 / 3}
ref_pauli2 = {'01': 2 / 9, '11': 4 / 9, '10': 2 / 9, '00': 1 / 9}
ref_depol1 = {'1': 1 / 2, '0': 1 / 2}
ref_depol2 = {'01': 1 / 4, '11': 1 / 4, '10': 1 / 4, '00': 1 / 4}
ref_cumul = {'0': 1/3, '1': 2/3}


def assert_freq_dict_almost_equal(d1, d2, atol):
    """ Utility function to check whether two frequency dictionaries are almost equal, for arbitrary tolerance """
    if d1.keys() != d2.keys():
        raise AssertionError("Dictionary keys differ. Frequency dictionaries are not almost equal.\n"
                             f"d1 keys: {d1.keys()} \nd2 keys: {d2.keys()}")
    else:
        for k in d1.keys():
            if abs(d1[k] - d2[k]) > atol:
                raise AssertionError(f"Dictionary entries beyond tolerance {atol}: \n{d1} \n{d2}")
    return True


class TestSimulate(unittest.TestCase):

    def test_noisy_simulation_not_supported(self):
        """
            Ensures that an error is returned if user attempts to run noisy simulation on a backend that does
            not support it as part of this package.
        """
        for b, s in backend_info.items():
            if not s['noisy_simulation']:
                self.assertRaises(ValueError, Simulator, target=b, n_shots=1, noise_model=True)

    def test_cannot_cumulate_same_noise_same_gate(self):
        """ Ensures an error is returned if user attempts to cumulate the same type of noise on the same gate """
        nm = NoiseModel()
        nm.add_quantum_error('X', 'depol', 0.2)
        self.assertRaises(ValueError, nm.add_quantum_error, 'X', 'depol', 0.3)

    def test_unsupported_noise_channel(self):
        """ Ensures an error is returned if user attempts to user an unsupported noise channel """
        nm = NoiseModel()
        self.assertRaises(ValueError, nm.add_quantum_error, 'X', 'incorrect_input', 0.3)

    def test_incorrect_arguments(self):
        """ Ensures an error is returned if noise parameters are incorrect """
        nm = NoiseModel()
        self.assertRaises(ValueError, nm.add_quantum_error, 'X', 'pauli', 0.3)
        self.assertRaises(ValueError, nm.add_quantum_error, 'X', 'depol', 1)
        self.assertRaises(ValueError, nm.add_quantum_error, 'X', 'depol', [0.3, 0.2, 0.1])

    # TODO: replace the noise channel name by one supported by agnostic simulator, but not with qiskit backend later
    @unittest.skipIf("qiskit" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_unsupported_noise_channel_qiskit(self):
        """ Ensures an error is returned if user attempts to user an unsupported noise channel, for qiskit """
        nm = NoiseModel()
        self.assertRaises(ValueError, nm.add_quantum_error, 'X', 'dummy', 0.3)

    @unittest.skipIf("qiskit" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_qiskit_noise_dictionary_rotations(self):
        """ Generate noise dictionary using qiskit gates as keys. Map rotation gates to U-gates with no redundancy
         Ensure results as expected."""

        nt, np = 'pauli', [0.5, 0.25, 0.25]
        nm = NoiseModel()
        for g in {"RX", "RY", "RZ"}:
            nm.add_quantum_error(g, nt, np)

        qnd = get_qiskit_noise_dict(nm)
        for g in {'u1', 'u2', 'u3'}:
            assert(g in qnd)
            assert(qnd[g] == [(nt, np)])

    @unittest.skipIf("qulacs" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_noisy_simulation_qulacs(self):
        """
            Test noisy simulation through qulacs.
            Currently tested: pauli noise, depolarization noise (1 and 2 qubit gates)
        """

        # Pauli noise for one- and two-qubit gates. Circuits are only a X gate, or just a CNOT gate.
        s_nmp = Simulator(target='qulacs', n_shots=10**6, noise_model=nmp)
        res_pauli1, _ = s_nmp.simulate(cn1)
        assert_freq_dict_almost_equal(res_pauli1, ref_pauli1, 1e-2)
        res_pauli2, _ = s_nmp.simulate(cn2)
        assert_freq_dict_almost_equal(res_pauli2, ref_pauli2, 1e-2)

        # Depol noise for one- and two-qubit gates. Circuits are only a X gate or just a CNOT gate.
        s_nmd = Simulator(target='qulacs', n_shots=10**6, noise_model=nmd)
        res_depol1, _ = s_nmd.simulate(cn1)
        assert_freq_dict_almost_equal(res_depol1, ref_depol1, 1e-2)
        res_depol2, _ = s_nmd.simulate(cn2)
        assert_freq_dict_almost_equal(res_depol2, ref_depol2, 1e-2)

        # Cumulate several noises on a given gate (here noise simplifies to identity)
        s_nmc = Simulator(target='qulacs', n_shots=10**6, noise_model=nmc)
        res_cumul, _ = s_nmc.simulate(cn1)
        assert_freq_dict_almost_equal(res_cumul, ref_cumul, 1e-2)

    @unittest.skipIf("qiskit" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_noisy_simulation_qiskit(self):
        """
            Test noisy simulation through qiskit.
            Currently tested: pauli noise, depolarization noise (1 and 2 qubit gates)
        """

        # Pauli noise for one- and two-qubit gates. Circuits are only a X gate, or just a CNOT gate.
        s_nmp = Simulator(target='qiskit', n_shots=10**6, noise_model=nmp)
        res_pauli1, _ = s_nmp.simulate(cn1)
        assert_freq_dict_almost_equal(res_pauli1, ref_pauli1, 1e-2)
        res_pauli2, _ = s_nmp.simulate(cn2)
        assert_freq_dict_almost_equal(res_pauli2, ref_pauli2, 1e-2)

        # Depol noise for one- and two-qubit gates. Circuits are only a X gate or just a CNOT gate.
        s_nmd = Simulator(target='qiskit', n_shots=10**6, noise_model=nmd)
        res_depol1, _ = s_nmd.simulate(cn1)
        assert_freq_dict_almost_equal(res_depol1, ref_depol1, 1e-2)
        res_depol2, _ = s_nmd.simulate(cn2)
        assert_freq_dict_almost_equal(res_depol2, ref_depol2, 1e-2)

        # Cumulate several noises on a given gate (here noise simplifies to identity)
        s_nmp = Simulator(target='qiskit', n_shots=10**6, noise_model=nmc)
        res_cumul, _ = s_nmp.simulate(cn1)
        assert_freq_dict_almost_equal(res_cumul, ref_cumul, 1e-2)

    @unittest.skipIf("cirq" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_noisy_simulation_cirq(self):
        """
            Test noisy simulation through cirq.
            Currently tested: pauli noise, depolarization noise (1 and 2 qubit gates)
        """

        # Pauli noise for one- and two-qubit gates. Circuits are only a X gate, or just a CNOT gate.
        s_nmp = Simulator(target='cirq', n_shots=10**6, noise_model=nmp)
        res_pauli1, _ = s_nmp.simulate(cn1)
        assert_freq_dict_almost_equal(res_pauli1, ref_pauli1, 1e-2)
        res_pauli2, _ = s_nmp.simulate(cn2)
        assert_freq_dict_almost_equal(res_pauli2, ref_pauli2, 1e-2)

        # Depol noise for one- and two-qubit gates. Circuits are only a X gate or just a CNOT gate.
        s_nmd = Simulator(target='cirq', n_shots=10**6, noise_model=nmd)
        res_depol1, _ = s_nmd.simulate(cn1)
        assert_freq_dict_almost_equal(res_depol1, ref_depol1, 1e-2)
        res_depol2, _ = s_nmd.simulate(cn2)
        assert_freq_dict_almost_equal(res_depol2, ref_depol2, 1e-2)

        # Cumulate several noises on a given gate (here noise simplifies to identity)
        s_nmc = Simulator(target='cirq', n_shots=10**6, noise_model=nmc)
        res_cumul, _ = s_nmc.simulate(cn1)
        assert_freq_dict_almost_equal(res_cumul, ref_cumul, 1e-2)

    def test_get_expectation_value_noisy(self):
        """Test of the get_expectation_value function with a noisy simulator"""
        # Test Hamiltonian.
        H = QubitOperator()
        H.terms = {(): -14.41806525945003, ((0, 'Z'),): 0.0809953994342687,
                   ((1, 'Z'),): 0.0809953994342687, ((0, 'Z'), (1, 'Z')): 0.0077184273651725865,
                   ((0, 'X'), (1, 'X')): 0.0758664717894615}

        # Hard coding of test circuit.
        circuit = Circuit()
        circuit.add_gate(Gate("RX", 0, parameter=3.141595416808))
        circuit.add_gate(Gate("RX", 1,  parameter=3.141588753134))
        circuit.add_gate(Gate("H", 0))
        circuit.add_gate(Gate("RX", 1, parameter=1.570796326795))
        circuit.add_gate(Gate("CNOT", 1, 0))
        circuit.add_gate(Gate("RZ", 1, parameter=0.43912793, is_variational=True))
        circuit.add_gate(Gate("CNOT", 1, 0))
        circuit.add_gate(Gate("RX", 1, parameter=10.995574287564))
        circuit.add_gate(Gate("H", 0))

        # No Noise model.
        nmp_no_noise = NoiseModel()
        noise = 0.00
        nmp_no_noise.add_quantum_error("CNOT", "pauli", [noise, noise, noise])
        sim_no_noise = Simulator(target=default_simulator, n_shots=10**6, noise_model=nmp_no_noise)

        # Small Noise model
        nmp_small_noise = NoiseModel()
        noise = 0.01
        nmp_small_noise.add_quantum_error("CNOT", "pauli", [noise, noise, noise])
        sim_small_noise = Simulator(target=default_simulator, n_shots=10**6, noise_model=nmp_small_noise)

        energy_no_noise = sim_no_noise.get_expectation_value(H, circuit)
        energy_small_noise = sim_small_noise.get_expectation_value(H, circuit)

        self.assertAlmostEqual(energy_no_noise, -14.58922316, delta=1e-2)
        self.assertAlmostEqual(energy_small_noise, -14.58922316, delta=1e-1)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    This python module stores template Q# strings that can be used to have Q# code files ready
    to be shared with collaborators, submitted to a remote compute backend through Microsoft services
    or compiled by the local QDK simulator.
"""

_header = '''namespace MyNamespace
{
    open Microsoft.Quantum.Intrinsic;
    open Microsoft.Quantum.Canon;
    open Microsoft.Quantum.Arrays;
    open Microsoft.Quantum.Extensions.Convert;
    open Microsoft.Quantum.Characterization;
    open Microsoft.Quantum.Measurement;\n
'''

_qdk_template = '''
    /// # Summary:
    ///         Returns the estimated probabilities associated to the different measurements
    ///         after applying the state-preparation provided by the user at runtime to the quantum state
    operation EstimateFrequencies(nQubits : Int, nShots : Int) : Double[]
    {{
        mutable frequencies = new Double[2^nQubits];

        for (iShot in 1..nShots)
        {{
            // Apply Q# operation (state preparation to measurements) to qubit register of size nQubits
            mutable results = {operation_name}();

            // Update frequencies based on sample value
            mutable index = 0;
            for (iQubit in nQubits-1..-1..0)
            {{
                if (results[iQubit] == One) {{set index = index + 2^iQubit;}}
            }}

            set frequencies w/= index <- frequencies[index] + 1.0;
        }}

        // Rescale to obtain frequencies observed
        for (iFreq in 0..2^nQubits-1)
        {{
            set frequencies w/= iFreq <- frequencies[iFreq] / ToDouble(nShots);
        }}

        return frequencies;
    }}

    /// INSERT TRANSLATED CIRCUIT HERE
'''



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np

from tangelo.linq import Gate, Circuit
from tangelo.toolboxes.measurements import RandomizedClassicalShadow
from tangelo.toolboxes.operators import QubitOperator

# Circuit to sample (Bell state).
state = Circuit([Gate("H", 0), Gate("CNOT", 1, 0)])

# Simple saved bistrings and unitaries to construct a shadow (size = 100).
bitstrings = ["01", "00", "00", "01", "00", "00", "11", "11", "00", "11", "11",
              "01", "00", "11", "11", "00", "11", "11", "00", "11", "11", "10",
              "01", "11", "01", "00", "00", "01", "01", "11", "01", "01", "11",
              "10", "00", "01", "01", "00", "00", "10", "10", "10", "00", "11",
              "00", "01", "00", "11", "00", "00", "11", "11", "11", "00", "10",
              "01", "01", "10", "01", "00", "00", "10", "00", "00", "10", "10",
              "01", "00", "11", "01", "00", "11", "11", "00", "11", "11", "01",
              "01", "01", "11", "00", "11", "10", "11", "10", "00", "00", "00",
              "00", "10", "01", "10", "10", "11", "11", "11", "01", "00", "11",
              "11"]
unitaries = ["ZY", "XZ", "XZ", "YZ", "ZX", "XX", "YZ", "XZ", "YX", "ZZ", "XX",
             "YY", "XX", "XX", "ZZ", "YZ", "XX", "XZ", "XY", "YX", "XZ", "ZY",
             "YX", "XY", "YZ", "XX", "XX", "ZY", "ZY", "XZ", "YY", "XZ", "YX",
             "YY", "ZZ", "YY", "YZ", "ZX", "XY", "XY", "YY", "YY", "XZ", "YZ",
             "XZ", "XZ", "ZZ", "XX", "XZ", "ZX", "ZZ", "ZX", "ZZ", "XX", "YX",
             "ZX", "XY", "YY", "YY", "XX", "YX", "YZ", "XX", "ZZ", "XZ", "YY",
             "YX", "ZY", "XZ", "ZX", "XX", "XX", "YX", "XY", "ZZ", "XZ", "ZX",
             "ZY", "YX", "ZZ", "XX", "ZY", "YZ", "ZX", "YZ", "ZY", "XZ", "XX",
             "XZ", "YZ", "YY", "ZY", "XZ", "XX", "ZY", "XX", "YY", "XY", "ZX",
             "ZX"]


class RandomizedClassicalShadowTest(unittest.TestCase):

    def test_initialization(self):
        """Testing the initialization."""

        RandomizedClassicalShadow(state, bitstrings, unitaries)

    def test_same_length_check(self):
        """Testing the case where arguments are not the same length."""

        wrong_bitstrings = bitstrings + ["00"]
        with self.assertRaises(AssertionError):
            RandomizedClassicalShadow(state, wrong_bitstrings, unitaries)

    def test_shadow_properties(self):
        """Testing of the shadow properties."""

        cs = RandomizedClassicalShadow(state, bitstrings, unitaries, shuffle=False)

        self.assertEqual(cs.n_qubits, 2)
        self.assertEqual(cs.size, 100)
        self.assertEqual(len(cs), 100)

    def test_get_term_observable(self):
        """Testing the computation of a single qubit term."""

        cs = RandomizedClassicalShadow(state, bitstrings, unitaries, shuffle=False)
        obs = cs.get_term_observable([(0, "Y"), (1, "Y")], 1., k=10)
        self.assertAlmostEqual(obs, -0.89999, places=4)

    def test_get_observable(self):
        """Testings the computation of an eigenvalue of a QubitOperator."""

        cs = RandomizedClassicalShadow(state, bitstrings, unitaries, shuffle=False)
        obs = cs.get_observable(QubitOperator("Y0 Y1", coefficient=1.))
        self.assertAlmostEqual(obs, -0.89999, places=4)

    def test_estimate_state(self):
        """Testing of the state estimation method to get the density matrix."""

        cs = RandomizedClassicalShadow(state, bitstrings, unitaries, shuffle=False)
        rho_estimate = cs.estimate_state()

        # Previously ran with this specific shadow.
        rho_ref = np.array([
            [ 0.445+0.j, 0.06+0.045j, 0.255+0.09j, 0.63-0.09j],
            [ 0.06-0.045j, 0.07+0.j, 0.135+0.045j, -0.15-0.09j],
            [ 0.255-0.09j, 0.135-0.045j, 0.025+0.j, -0.12-0.045j],
            [ 0.63+0.09j, -0.15+0.09j, -0.12+0.045j, 0.46+0.j]
        ])

        np.testing.assert_array_almost_equal(rho_estimate, rho_ref)

    def test_get_basis_circuits(self):
        """Testing of the method to get the appended circuit corresponding to
        the unitaries.
        """
        cs = RandomizedClassicalShadow(state, bitstrings, unitaries)

        self.assertEqual(len(cs.get_basis_circuits(False)), 100)
        self.assertEqual(len(cs.get_basis_circuits(True)), 9)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os

from tangelo import SecondQuantizedMolecule
from tangelo.molecule_library import mol_H2_sto3g, xyz_H2O
from tangelo.toolboxes.molecular_computation.molecule import atom_string_to_list


H2_list = [("H", (0., 0., 0.)), ("H", (0., 0., 0.7414))]
H2_string = """
H       0.0        0.0        0.0
H       0.0        0.0        0.7414
"""
H2_file_xyz = os.path.dirname(os.path.abspath(__file__))+"/data/h2.xyz"

H2O_list = [("O", (0., 0., 0.11779)),
            ("H", (0., 0.75545, -0.47116)),
            ("H", (0., -0.75545, -0.47116))]


def atom_list_close(atom1, atom2, atol):

    for (a0, xyz0), (a1, xyz1) in zip(atom1, atom2):
        if a0 != a1:
            raise AssertionError(f"Atoms are not the same {a0} != {a1}")
        for x0, x1 in zip(xyz0, xyz1):
            if abs(x0 - x1) > atol:
                raise AssertionError(f"geometries for atom {a0} are different. {x0} != {x1}")


class CoordsTest(unittest.TestCase):

    def test_atoms_string_to_list(self):
        """Verify conversion from string to list format for atom coordinates."""
        assert(atom_string_to_list(H2_string) == H2_list)


class SecondQuantizedMoleculeTest(unittest.TestCase):

    def test_instantiate_H2(self):
        """Verify basic properties of molecule object through instantiation of
        MolecularData class.
        """

        molecule = SecondQuantizedMolecule(H2_list, 0, 0, "sto-3g")

        assert(molecule.elements == ["H"]*2)
        assert(molecule.basis == "sto-3g")
        assert(molecule.spin == 0)
        assert(molecule.q == 0)
        assert(molecule.n_electrons == 2)

        molecule = SecondQuantizedMolecule(H2_file_xyz, 0, 0, "sto-3g")
        assert(molecule.elements == ["H"]*2)
        assert(molecule.basis == "sto-3g")
        assert(molecule.spin == 0)
        assert(molecule.q == 0)
        assert(molecule.n_electrons == 2)
        atom_list_close(molecule.xyz, H2_list, 1.e-14)

    def test_freezing_orbitals(self):
        """Verify freezing orbitals functionalities."""

        no_freezing = SecondQuantizedMolecule(H2O_list, 0, 0, "sto-3g", frozen_orbitals=None)
        assert(no_freezing.active_occupied == [0, 1, 2, 3, 4])
        assert(no_freezing.active_virtual == [5, 6])
        assert(no_freezing.frozen_occupied == [])
        assert(no_freezing.frozen_virtual == [])

        freeze_with_int = SecondQuantizedMolecule(H2O_list, frozen_orbitals=3)
        assert(freeze_with_int.active_occupied == [3, 4])
        assert(freeze_with_int.active_virtual == [5, 6])
        assert(freeze_with_int.frozen_occupied == [0, 1, 2])
        assert(freeze_with_int.frozen_virtual == [])

        freeze_with_list = SecondQuantizedMolecule(H2O_list, frozen_orbitals=[0, 1, 2, 6])
        assert(freeze_with_list.active_occupied == [3, 4])
        assert(freeze_with_list.active_virtual == [5])
        assert(freeze_with_list.frozen_occupied == [0, 1, 2])
        assert(freeze_with_list.frozen_virtual == [6])

    def test_freezing_empty(self):
        """Verify freezing orbitals empty input."""

        # None should result in the same as nothing.
        none_as_frozen = SecondQuantizedMolecule(H2O_list, frozen_orbitals=None)
        assert(none_as_frozen.active_occupied == [0, 1, 2, 3, 4])
        assert(none_as_frozen.active_virtual == [5, 6])
        assert(none_as_frozen.frozen_occupied == [])
        assert(none_as_frozen.frozen_virtual == [])

        # An empty list should result in the same as nothing.
        empty_as_frozen = SecondQuantizedMolecule(H2O_list, frozen_orbitals=[])
        assert(empty_as_frozen.active_occupied == [0, 1, 2, 3, 4])
        assert(empty_as_frozen.active_virtual == [5, 6])
        assert(empty_as_frozen.frozen_occupied == [])
        assert(empty_as_frozen.frozen_virtual == [])

    def test_freezing_type_exception(self):
        """Verify freezing orbitals exceptions."""

        # Cases where the input is wrong type.
        with self.assertRaises(TypeError):
            SecondQuantizedMolecule(H2O_list, frozen_orbitals="3")
        with self.assertRaises(TypeError):
            SecondQuantizedMolecule(H2O_list, frozen_orbitals=3.141592)
        with self.assertRaises(TypeError):
            SecondQuantizedMolecule(H2O_list, frozen_orbitals=[0, 1, 2.2222, 3, 4, 5])

    def test_no_active_electron(self):
        """Verify if freezing all active orbitals fails."""

        # Cases where no active electron are remaining.
        with self.assertRaises(ValueError):
            SecondQuantizedMolecule(H2O_list, frozen_orbitals=5)
        with self.assertRaises(ValueError):
            SecondQuantizedMolecule(H2O_list, frozen_orbitals=[0, 1, 2, 3, 4, 5])

    def test_no_active_virtual(self):
        """Verify if freezing all virtual orbitals fails."""

        # Cases where no active virtual orbitals are remaining.
        with self.assertRaises(ValueError):
            SecondQuantizedMolecule(H2O_list, frozen_orbitals=[5, 6])

    def test_all_active_orbitals_occupied_but_some_not_fully(self):
        """Verify that having all active orbitals occupied but only partially occupied is permitted"""
        try:
            SecondQuantizedMolecule(H2O_list, frozen_orbitals=[0, 1, 2, 3, 6], spin=2)
        except ValueError:
            self.fail("Unexpected ValueError raised")

    def test_get_fermionic_hamiltonian(self):
        """Verify energy shift in molecular hamiltonian."""

        molecule = SecondQuantizedMolecule(H2O_list, frozen_orbitals=1)
        shift = molecule.fermionic_hamiltonian.constant
        self.assertAlmostEqual(shift, -51.47120372466002, delta=1e-6)

    def test_energy_from_rdms(self):
        """Verify energy computation from RDMs."""

        rdm1 = [[ 1.97453997e+00, -7.05987336e-17],
                [-7.05987336e-17,  2.54600303e-02]]

        rdm2 = [
            [[[ 1.97453997e+00, -7.96423130e-17],
              [-7.96423130e-17,  3.21234218e-33]],
             [[-7.96423130e-17, -2.24213843e-01],
              [ 0.00000000e+00,  9.04357944e-18]]],
            [[[-7.96423130e-17,  0.00000000e+00],
              [-2.24213843e-01,  9.04357944e-18]],
             [[ 3.21234218e-33,  9.04357944e-18],
              [ 9.04357944e-18,  2.54600303e-02]]]
            ]

        e_rdms = mol_H2_sto3g.energy_from_rdms(rdm1, rdm2)
        self.assertAlmostEqual(e_rdms, -1.1372701, delta=1e-5)

    def test_symmetry_label(self):
        """Verify that the symmetry labels are correct when symmetry=True or symmetry="C2v"."""
        mo_symm_labels = ["A1", "A1", "B1", "A1", "B2", "A1", "B1"]
        mo_symm_ids = [0, 0, 2, 0, 3, 0, 2]

        molecule = SecondQuantizedMolecule(xyz=xyz_H2O, q=0, spin=0, symmetry=True, basis="sto-3g")
        assert(mo_symm_labels == molecule.mo_symm_labels)
        assert(mo_symm_ids == molecule.mo_symm_ids)

        molecule = SecondQuantizedMolecule(xyz=xyz_H2O, q=0, spin=0, symmetry="C2v", basis="sto-3g")
        assert(mo_symm_labels == molecule.mo_symm_labels)
        assert(mo_symm_ids == molecule.mo_symm_ids)

    def test_ecp(self):
        """Verify that the number of electrons is reduced when ecp is called."""

        molecule = SecondQuantizedMolecule(xyz="Yb", q=0, spin=0, basis="crenbl", ecp="crenbl")
        # "Yb" has 70 electrons but the ecp reduces this to 16
        assert(molecule.n_active_electrons == 16)
        assert(molecule.n_active_mos == 96)

        molecule = SecondQuantizedMolecule(xyz="Cu", q=0, spin=1, basis="cc-pvdz", ecp="crenbl",
                                           frozen_orbitals=list(range(8)))
        # "Cu" has 29 electrons but the ecp reduces this to 19. The active electrons are 19 - 8 * 2 = 3
        assert(molecule.n_active_electrons == 3)
        assert(molecule.n_active_mos == 35)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for data structure for DMET fragments."""

from dataclasses import dataclass, field
import openfermion
import numpy as np
import pyscf
from pyscf import ao2mo

from tangelo.toolboxes.operators import FermionOperator
from tangelo.toolboxes.qubit_mappings.mapping_transform import get_fermion_operator


@dataclass
class SecondQuantizedDMETFragment:
    """Mimicking SecondQuantizedMolecule for DMET fragments. It has the minimal
    number of attributes and methods to be parsed by electronic solvers.
    """

    molecule: pyscf.gto
    mean_field: pyscf.scf

    # Fragment data computed by the DMET backend code. Useful when computing the
    # energy (converting embedded problem to full system energy).
    fock: np.array
    fock_frag_copy: np.array
    t_list: list
    one_ele: np.array
    two_ele: np.array

    n_active_electrons: int = field(init=False)
    n_active_sos: int = field(init=False)
    q: int = field(init=False)
    spin: int = field(init=False)

    basis: str = field(init=False)
    n_active_mos: int = field(init=False)
    fermionic_hamiltonian: FermionOperator = field(init=False, repr=False)
    frozen_mos: None = field(init=False)

    def __post_init__(self):
        self.n_active_electrons = self.molecule.nelectron
        self.q = self.molecule.charge
        self.spin = self.molecule.spin

        self.basis = self.molecule.basis
        self.n_active_mos = len(self.mean_field.mo_energy)
        self.n_active_sos = 2*self.n_active_mos

        self.fermionic_hamiltonian = self._get_fermionic_hamiltonian()
        self.frozen_mos = None

    def _get_fermionic_hamiltonian(self):
        """This method returns the fermionic hamiltonian. It written to take
        into account calls for this function is without argument, and attributes
        are parsed into it.

        Returns:
            FermionOperator: Self-explanatory.
        """

        dummy_of_molecule = openfermion.MolecularData([["C", (0., 0., 0.)]], "sto-3g", self.spin+1, self.q)

        # Overwrting nuclear repulsion term.
        dummy_of_molecule.nuclear_repulsion = self.mean_field.mol.energy_nuc()

        canonical_orbitals = self.mean_field.mo_coeff
        h_core = self.mean_field.get_hcore()
        n_orbitals = len(self.mean_field.mo_energy)

        # Overwriting 1-electron integrals.
        dummy_of_molecule._one_body_integrals = canonical_orbitals.T @ h_core @ canonical_orbitals

        twoint = self.mean_field._eri
        eri = ao2mo.restore(8, twoint, n_orbitals)
        eri = ao2mo.incore.full(eri, canonical_orbitals)
        eri = ao2mo.restore(1, eri, n_orbitals)

        # Overwriting 2-electrons integrals.
        dummy_of_molecule._two_body_integrals = np.asarray(eri.transpose(0, 2, 3, 1), order="C")

        fragment_hamiltonian = dummy_of_molecule.get_molecular_hamiltonian()

        return get_fermion_operator(fragment_hamiltonian)

    def to_pyscf(self, basis=None):
        """Method to output the PySCF molecule.

        Returns
            pyscf.gto.Mole: PySCF molecule.
        """
        return self.molecule



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides information about what the backends currently installed in
the user's environment, for the purpose of running / skipping tests, and setting
a default simulator.
"""

import os
import sys


class HiddenPrints:
    """Class to hide terminal printing with a 'with' block."""

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def is_package_installed(package_name):
    try:
        exec(f"import {package_name}")
        # DEBUG print(f'{package_name:16s} :: found')
        return True
    except ModuleNotFoundError:
        # DEBUG print(f'{package_name:16s} :: not found')
        return False


# List all backends and statevector ones supported by Simulator class
all_backends = {"qulacs", "qiskit", "cirq", "braket", "projectq", "qdk"}
all_backends_simulator = {"qulacs", "qiskit", "cirq", "qdk"}
sv_backends_simulator = {"qulacs", "qiskit", "cirq"}

# Dictionary mapping package names to their identifier in this module
packages = {p: p for p in all_backends}
packages["qdk"] = "qsharp"

# Figure out what is install in user's environment
installed_backends = {p_id for p_id, p_name in packages.items() if is_package_installed(p_name)}
installed_simulator = installed_backends & all_backends_simulator
installed_sv_simulator = installed_backends & sv_backends_simulator

# Check if qulacs installed (best performance for tests). If not, defaults to cirq (always installed with openfermion)
default_simulator = "qulacs" if "qulacs" in installed_sv_simulator else "cirq"



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides helpers in order to import data coming from a MI-FNO job
run on QEMIST Cloud, providing the users with both fragment information as well as
reference results obtained by the classical solvers in QEMIST Cloud. The
fragments can be passed to a quantum solver or be used for a quantum computing
experiment.

Currently, the fragment energies can only be recomputed with a quantum
algorihtms (the interface between MI-FNO fragments and classical algorithms is not
implemented yet).
"""

from functools import reduce
import itertools
import os
import requests
import warnings
import json

import h5py
import numpy as np
import pandas as pd


class MIFNOHelper():
    """Python object to post-process, fetch and manipulate QEMIST Cloud MI-FNO
    results. The use case for this is to map MI-FNO subproblems to
    fermionic Hamiltonians acting as inputs. This object keeps track of the
    classical calculation results.

    Attributes:
        e_tot (float): Total MI-FNO energy.
        e_corr (float): Correlation energy (e_tot - e_mf).
        e_mf (float): Mean-field energy.
        frag_info (dict): Information about each fragment. The keys are related
            to the truncation number (int) . The nested dictionaries have keys
             refering to the sampled active space (e.g. '(1,)' or '(0, 2)') They
            contain information about the correction term, epsilon, list of
            truncated orbitals and more.
        verbose (bool): To print or not to print warnings.

    Properties:
        to_dataframe (pandas.DataFrame): Converted frag_info dict into a pandas
            DataFrame.
        fragment_ids (list of string): List of all fragment identifiers.
        frag_info_flattened (dictionary): The nested frag_info without the first
            layer (keys = truncation number).
    """

    def __init__(self, mi_json_file=None, fno_json_folder=None, mi_dict=None,
                 fno_dicts=None, verbose=False):
        """Initialization method to process the classical results. A json path
        or a python dictionary can be passed to the method for the MI
        or each FNO fragment results. Passing both a path and a dictionary
        raises an error. Not all fragment results need to be imported:
        in case of missing data, MIFNOHelper raises an error
        mentionning the missing pieces.

        Args:
            mi_json_file (string): Path to a json file containing the MI results
                from QEMIST Cloud.
            fno_json_folder (string): Path to a folder containing the FNO
                fragment (json files) results from QEMIST-Cloud.
            mi_dict (dict): MI results (QEMIST Cloud output).
            fno_dicts (list of dicts): FNO fragment results (QEMIST Cloud
                output).
            verbose (bool): Verbosity.
        """

        # Raise error/warnings if input is not as expected. Only a single input
        # must be provided to avoid conflicts.
        if not (bool(mi_json_file) ^ bool(mi_dict)):
            raise ValueError(f"A json file path OR a dictionary object for MI \
                               results must be provided when instantiating \
                               {self.__class__.__name__}.")

        if not (bool(fno_json_folder) ^ bool(fno_dicts)):
            raise ValueError(f"A json folder path OR a dictionary object for \
                               FNO fragment results must be provided when \
                               instantiating {self.__class__.__name__}.")

        self.verbose = verbose

        if mi_json_file:
            assert os.path.isfile(mi_json_file), f"The file {mi_json_file} does not exist."

            with open(mi_json_file, "r") as f:
                mi_dict = json.loads(f.read())

            mi_dict["subproblem_data"] = {int(k): v for k, v in mi_dict["subproblem_data"].items()}

        # Incremental (problem decomposition) quantities.
        self.e_tot = mi_dict["energy_total"]
        self.e_corr = mi_dict["energy_correlation"]
        self.e_mf = mi_dict["mean_field_energy"]

        # Select only the MI relevant keys.
        self.frag_info = dict()
        for n_body, fragments_per_truncation in mi_dict["subproblem_data"].items():
            self.frag_info[n_body] = dict()
            for frag_id, frag_result in fragments_per_truncation.items():
                if frag_result.get("problem_handle", None) is not None:
                    self.frag_info[n_body][frag_id] = {k: frag_result.get(k, None) for k in {"epsilon", "problem_handle"}}

        # Read fragment results stored in json files in a specified folder.
        if fno_json_folder:
            absolute_path = os.path.abspath(fno_json_folder)

            fno_dicts = list()
            for file in os.listdir(absolute_path):
                if file.endswith(".json"):
                    with open(os.path.join(absolute_path, file), "r") as f:
                        frag_results = json.loads(f.read())
                    fno_dicts.append(frag_results)

        fragment_relevant_info = {
            "energy_total",
            "energy_correlation",
            "correction",
            "frozen_orbitals_truncated",
            "complete_orbital_space",
            "mo_coefficients"
        }

        # Clean FNO fragment results.
        for frag_id in self.fragment_ids:
            n_body = len(eval(frag_id))

            # Default value if fragment information is not detected.
            frag_results = dict()

            for fno_frag in fno_dicts:
                # Fragments are identified by the active occupied orbital space,
                # or their unique problem_handle.
                if str(tuple(fno_frag["active_occupied_orbital_space"])) == frag_id:

                    # If the mean-field energy for the fragment is not the same,
                    # the coordinates, spin, basis or charge of the problem
                    # might be different.
                    assert round(self.e_mf, 6) == round(fno_frag["mean_field_energy"], 6), \
                        f"The mean-field energy for fragment {frag_id} is " \
                        "different than the one detected in the MI results."

                    # The mo_coefficients data are essential to recompute
                    # fermionic operators.
                    assert "mo_coefficients" in fno_frag, "MO coefficients "\
                        f"not found in the {frag_id} results. Verify that " \
                        "the `export_fragment_data` flag is set to True for " \
                        "the MI-FNO calculation in QEMIST Cloud."

                    frag_results = fno_frag

            self.frag_info[n_body][frag_id].update({k: frag_results.get(k, None) for k in fragment_relevant_info})

    def __repr__(self):
        """Format the object to print the energies and the fragment information
        as a pandas.DataFrame.
        """

        data_fragments = self.to_dataframe
        data_fragments.drop(["frozen_orbitals_truncated"], axis=1, inplace=True)
        data_fragments.drop(["complete_orbital_space"], axis=1, inplace=True)
        str_rep = f"(All the energy values are in hartree)\n" \
                  f"Total MI-FNO energy = {self.e_tot}\n" \
                  f"Correlation energy = {self.e_corr}\n" \
                  f"Mean-field energy = {self.e_mf}\n" \
                  f"{data_fragments}"

        return str_rep

    def __getitem__(self, frag_id):
        """The user can select the fragment information (python dictionary) with
        the [] operator.
        """
        return self.frag_info_flattened[frag_id]

    @property
    def to_dataframe(self):
        """Outputs the fragment informations as a pandas.DataFrame."""
        df = pd.DataFrame.from_dict(self.frag_info_flattened, orient="index")

        # Replace frozen_orbitals_truncated=None with an empty list.
        df["frozen_orbitals_truncated"] = df["frozen_orbitals_truncated"].apply(lambda d: d if isinstance(d, list) else [])
        df["complete_orbital_space"] = df["complete_orbital_space"].apply(lambda d: d if isinstance(d, list) else [])

        return df.drop(["mo_coefficients"], axis=1)

    @property
    def fragment_ids(self):
        """Output the fragment ids in a list."""
        return list(itertools.chain.from_iterable([d.keys() for d in self.frag_info.values()]))

    @property
    def frag_info_flattened(self):
        """Output the nested frag_info without the first layer."""
        return reduce(lambda a, b: {**a, **b}, self.frag_info.values())

    def retrieve_mo_coeff(self, destination_folder=os.getcwd()):
        """Function to fetch molecular orbital coefficients. A download path can
        be provided to change the directory where the files will be downloaded.
        If the files already exist, the function skips the download step. The
        array is stored in the ["mo_coefficients"]["array"] entry in the
        frag_info dictionary attribute.

        Args:
            destination_folder (string): Users can specify a path to a
                destination folder, where the files containing the coefficients
                will be downloaded. The default value is the directory where the
                user's python script is run.
        """
        if not os.path.isdir(destination_folder):
            raise FileNotFoundError(f"The {destination_folder} path does not exist.")
        absolute_path = os.path.abspath(destination_folder)

        # For each fragment, fetch the molecular orbital coefficients from the
        # HDF5 files.
        n_files = len(self.fragment_ids)
        i_file = 1
        for n_body_fragments in self.frag_info.values():
            for frag_id, frag in n_body_fragments.items():
                if frag.get("mo_coefficients", None):
                    file_path = os.path.join(absolute_path, frag["mo_coefficients"]["key"] + ".hdf5")

                    if not os.path.exists(file_path):
                        print(f"Downloading and writing MO coefficients file to {file_path} ({i_file} / {n_files})")
                        response = requests.get(frag["mo_coefficients"]["s3_url"])

                        with open(file_path, "wb") as file:
                            file.write(response.content)

                    with h5py.File(file_path, "r") as file:
                        mo_coeff = np.array(file["mo_coefficients"])

                    n_body_fragments[frag_id]["mo_coefficients"]["array"] = mo_coeff
                else:
                    print(f"MO coefficients for fragment {frag_id} ({i_file} / {n_files}) not available.")

                i_file += 1

    def compute_fermionoperator(self, molecule, frag_id):
        """Compute the fermionic Hamiltonian for a MI-FNO fragment.

        Args:
            molecule (SecondQuantizedMolecule): Full molecule description.
            frag_id (string): Fragment id, e.g. "(0, )", "(1, 2)", ...

        Returns:
            FermionOperator: Fermionic operator for the specified fragment id.
        """
        n_body = len(eval(frag_id))

        if self.frag_info[n_body][frag_id]["mo_coefficients"] is None:
            raise RuntimeError(f"The fragment information has not been imported.")

        if self.frag_info[n_body][frag_id]["mo_coefficients"].get("array", None) is None:
            raise RuntimeError(f"The molecular orbital coefficients are not available. Please call the {self.__class__.__name__}.retrieve_mo_coeff method.")

        mo_coeff = self.frag_info[n_body][frag_id]["mo_coefficients"]["array"]
        frozen_orbitals = self.frag_info[n_body][frag_id]["frozen_orbitals_truncated"]

        # Something is wrong if the molecule provided does not have the same
        # mean-field energy.
        assert round(molecule.mf_energy, 6) == round(self.e_mf, 6),  \
            "The molecule's mean-field energy is different than the one from " \
            "the results. Please verify that the molecular quantities are "\
            "the same as the one in the MI-FNO computation."

        # Returning a new molecule with the frozen orbitals.
        try:
            new_molecule = molecule.freeze_mos(frozen_orbitals, inplace=False)
        except ValueError:
            raise ValueError(f"All orbitals except {frag_id} are frozen from "
                             "the FNO truncation. That means no "
                             "correlation energy can be extracted from this "
                             "fragment.")

        return new_molecule._get_fermionic_hamiltonian(mo_coeff)

    def mi_summation(self, user_provided_energies=None, force_negative_epsilon=False):
        r"""Recompute the total energy for the method of increments (MI).
        Each increment corresponds to "new" correlation energy from the n-body
        problem. This method makes computing the total energy with new
        results possible.

        It computes the epsilons with the MP2 correction:
        \epsilon_{i} = E_c(i)
        \epsilon_{ij} = E_c(ij) - \epsilon_{i} - \epsilon_{i}
        \epsilon_{ijk} = E_c(ijk) - \epsilon_{ij} - \epsilon_{ik}
            - \epsilon_{jk} - \epsilon_{i} - \epsilon_{j} - \epsilon_{k}
        etc.

        Args:
            user_provided_energies (dict): New energy values provided by the
                user, used instead of the corresponding pre-computed ones. E.g.
                {"(0, )": -1.234} or {"(1, )": -1.234, "(0, 1)": -5.678}.
            force_negative_epsilon (bool): Force positive epsilons to 0.

        Returns:
            float: Method of increment total energy.
        """
        fragment_energies = {k: v["energy_total"] for k, v in self.frag_info_flattened.items()}

        if any([e is None for e in fragment_energies.values()]):
            raise ValueError("All fragment data must be imported to "
                "recompute the total MI-FNO energy.")

        if user_provided_energies is None:
            user_provided_energies = dict()
        else:
            fragment_correction = {k: v["correction"] for k, v in self.frag_info_flattened.items()}

            if any(fragment_correction[frag_id] is None for frag_id in user_provided_energies.keys()):
                raise RuntimeError(f"Not all the fragments in {list(user_provided_energies.keys())} "
                    "have been imported. The MP2 correction must be known "
                    "for all fragments to recompute the total MI-FNO energy.")

            user_provided_energies = {frag_id: e + fragment_correction[frag_id] for frag_id, e in user_provided_energies.items()}

        # Update to consider energies taken from a calculation.
        fragment_energies.update(user_provided_energies)

        # Equivalent to truncation_order in QEMIST Cloud.
        n_body_max = max(self.frag_info.keys())

        # Perform the incremental sumamtion.
        epsilons = dict()
        for n_body in range(1, n_body_max + 1):
            for frag_id in self.frag_info[n_body].keys():
                corr_energy = fragment_energies[frag_id] - self.e_mf
                epsilons[frag_id] = corr_energy

                if n_body > 1:
                    for n_increment in range(1, n_body):
                        for frag_increment in itertools.combinations(eval(frag_id), n_increment):
                            epsilons[frag_id] -= epsilons[str(frag_increment)]

        # Check if epsilon > 0, i.e. positive correlation energy increment.
        for frag_id, eps in epsilons.items():
            if eps > 0.:
                if self.verbose:
                    warnings.warn(f"Epsilon for frag_id {frag_id} is positive "
                        f"({eps}). With MI, there is no reason to consider a "
                        "fragment returning a positive correlation energy. "
                        "Please check your calculations.", RuntimeWarning)
                if force_negative_epsilon:
                    epsilons[frag_id] = 0.

        return self.e_mf + sum(epsilons.values())

    def n_electrons_spinorbs(self, frag_id):
        """Output the number of electrons and spinorbitals for a given fragment.

        Args:
            frag_id (string): Fragment id, e.g. "(0, )", "(1, 2)", ...

        Returns:
            int, int: Number of electrons, number of spinorbitals.
        """

        fragment_info = self.frag_info_flattened[frag_id]

        n_electrons = 2 * len(eval(frag_id))
        n_spinorbs = 2 * (len(fragment_info["complete_orbital_space"]) - len(fragment_info["frozen_orbitals_truncated"]))

        return n_electrons, n_spinorbs



# Copyright 2021 1QB Information Technologies Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing functions to manipulate 1- and 2-RDMs."""

import numpy as np


def matricize_2rdm(two_rdm, n_orbitals):
    """Turns the two_rdm tensor into a matrix for test purposes."""

    l = 0
    sq = n_orbitals * n_orbitals
    jpqrs = np.zeros((n_orbitals, n_orbitals), dtype=int)
    for i in range(n_orbitals):
        for j in range(n_orbitals):
            jpqrs[i, j] = l
            l += 1

    rho = np.zeros((sq, sq), dtype=complex)
    for i in range(n_orbitals):
        for j in range(n_orbitals):
            ij = jpqrs[i, j]
            for k in range(n_orbitals):
                for l in range(n_orbitals):
                    kl = jpqrs[k, l]
                    rho[ij, kl] += two_rdm[i, k, j, l]
    return rho



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from copy import copy

from tangelo.molecule_library import mol_H4_minao
from tangelo.problem_decomposition.dmet.dmet_problem_decomposition import Localization, DMETProblemDecomposition


class DMETVQETest(unittest.TestCase):

    def test_h4ring_vqe_uccsd(self):
        """DMET on H4 ring with fragment size one, using VQE-UCCSD."""

        opt_dmet = {"molecule": mol_H4_minao,
                    "fragment_atoms": [1, 1, 1, 1],
                    "fragment_solvers": ["vqe", "ccsd", "ccsd", "ccsd"],
                    "electron_localization": Localization.meta_lowdin,
                    "verbose": False
                    }

        # Run DMET
        dmet = DMETProblemDecomposition(opt_dmet)
        dmet.build()
        energy = dmet.simulate()

        # Test boostrapping error
        bootstrap_energy, standard_deviation = dmet.energy_error_bars(n_shots=50000, n_resamples=10, purify=False)
        rdm_measurements = copy(dmet.rdm_measurements)
        be_using_measurements, sd_using_measurements = dmet.energy_error_bars(n_shots=50000,
                                                                              n_resamples=10,
                                                                              purify=False,
                                                                              rdm_measurements=rdm_measurements)

        self.assertAlmostEqual(energy, -1.9916120594, delta=1e-3)
        # Should pass 99.993666% of the time with 4\sigma distance
        self.assertAlmostEqual(bootstrap_energy, -1.9916120594, delta=standard_deviation*4)
        self.assertAlmostEqual(be_using_measurements, -1.9916120594, delta=sd_using_measurements*4)

    def test_h4ring_vqe_resources(self):
        """Resources estimation on H4 ring."""

        opt_dmet = {"molecule": mol_H4_minao,
                    "fragment_atoms": [1, 1, 1, 1],
                    "fragment_solvers": ["vqe", "ccsd", "ccsd", "ccsd"],
                    "electron_localization": Localization.meta_lowdin,
                    "verbose": False
                    }

        # Building DMET fragments (with JW).
        dmet = DMETProblemDecomposition(opt_dmet)
        dmet.build()
        resources_jw = dmet.get_resources()

        # Building DMET fragments (with scBK).
        opt_dmet["solvers_options"] = {"qubit_mapping": "scbk", "initial_var_params": "ones", "up_then_down": True}
        dmet = DMETProblemDecomposition(opt_dmet)
        dmet.build()
        resources_bk = dmet.get_resources()

        # JW.
        self.assertEqual(resources_jw[0]["qubit_hamiltonian_terms"], 15)
        self.assertEqual(resources_jw[0]["circuit_width"], 4)
        # scBK.
        self.assertEqual(resources_bk[0]["qubit_hamiltonian_terms"], 5)
        self.assertEqual(resources_bk[0]["circuit_width"], 2)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np

from tangelo.linq import Simulator
from tangelo.linq.helpers.circuits import StateVector
from tangelo.helpers.utils import installed_backends


class StateVectorTest(unittest.TestCase):

    def test_init(self):
        """Test initialization of the ansatz class."""
        n_qubits = 3
        v = np.full((2**n_qubits), 1.+1j)
        v /= np.linalg.norm(v)

        # Test raises ValueError for vector of length not equal to 2**(integer)
        self.assertRaises(ValueError, StateVector, v[0:7])
        # Test raises ValueError if order does is not "msq_first" or "lsq_first"
        self.assertRaises(ValueError, StateVector, v, "not_msq_first_or_lsq_first")

    def test_circuits_representations_lsq(self):
        """Test initializing and uncomputing circuits with cirq lsq_first order"""
        sim = Simulator("cirq")
        v = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]) + 1j*np.array([0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1])
        v /= np.linalg.norm(v)

        sv = StateVector(v, order=sim.statevector_order)

        init_circ, phase = sv.initializing_circuit(return_phase=True)
        _, nsv = sim.simulate(init_circ, return_statevector=True)
        np.testing.assert_array_almost_equal(nsv*np.exp(1j*phase), v)

        uncomp_circ, phase = sv.uncomputing_circuit(return_phase=True)
        zero_state = np.zeros(8)
        zero_state[0] = 1
        _, nsv = sim.simulate(uncomp_circ, initial_statevector=v, return_statevector=True)
        np.testing.assert_array_almost_equal(nsv*np.exp(1j*phase), zero_state)

    @unittest.skipIf("qulacs" not in installed_backends, "Test Skipped: Backend not available \n")
    def test_circuits_representations_msq(self):
        """Test initializing and uncomputing circuits with qulacs msq_first order"""
        sim = Simulator("qulacs")
        v = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]) + 1j*np.array([0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1])
        v /= np.linalg.norm(v)

        sv = StateVector(v, order=sim.statevector_order)

        init_circ, phase = sv.initializing_circuit(return_phase=True)
        _, nsv = sim.simulate(init_circ, return_statevector=True)
        np.testing.assert_array_almost_equal(nsv*np.exp(1j*phase), v)

        uncomp_circ, phase = sv.uncomputing_circuit(return_phase=True)
        zero_state = np.zeros(8)
        zero_state[0] = 1
        _, nsv = sim.simulate(uncomp_circ, initial_statevector=v, return_statevector=True)
        np.testing.assert_array_almost_equal(nsv*np.exp(1j*phase), zero_state)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the Variationally Scheduled Quantum Simulation class. It provides an
Adiabatic State Preparation (ASP) inspired ansatz as described in https://arxiv.org/abs/2003.09913."""

import numpy as np
from openfermion import QubitOperator as ofQubitOperator

from .ansatz import Ansatz
from .ansatz_utils import get_exponentiated_qubit_operator_circuit
from tangelo.toolboxes.operators import FermionOperator
from tangelo.linq import Circuit
from tangelo.toolboxes.qubit_mappings.mapping_transform import get_qubit_number, fermion_to_qubit_mapping
from tangelo.toolboxes.qubit_mappings.statevector_mapping import get_reference_circuit


class VSQS(Ansatz):
    """This class implements the Variationally Scheduled Quantum Simulator (VSQS) for state preparation as described in
    https://arxiv.org/abs/2003.09913

    Must supply either a molecule or a qubit_hamiltonian. If supplying a qubit_hamiltonian, must also supply
    a reference_state Circuit and a h_init QubitOperator.

    Args:
        molecule (SecondQuantizedMolecule): The molecular system. Default: None
        mapping (str): One of the supported fermion to qubit mappings. Default : "JW"
        up_then_down (bool): Change basis ordering, putting all spin up orbitals first, followed by all spin down.
            Default: False (alternating spin up/down ordering)
        intervals (int): The number of steps in the VSQS process. Must be greater than 1. Default: 2
        time (float): The propagation time. Default: 1.
        qubit_hamiltonian (QubitOperator): The qubit Hamiltonian to evolve. Default: None
        reference_state (Circuit): The reference state for the propagation as defined by a Circuit. Mandatory if supplying
            a qubit_hamiltonian. Default: None
        h_init (QubitOperator): The initial qubit Hamiltonian that corresponds to the reference state. Mandatory if supplying
            a qubit_hamiltonian. Default: None
        h_nav (QubitOperator): The navigator Hamiltonian. Default: None
        trotter_order (int): The order of the Trotterization for each qubit operator. Default: 1
    """

    def __init__(self, molecule=None, mapping="jw", up_then_down=False, intervals=2, time=1., qubit_hamiltonian=None,
                 reference_state=None, h_init=None, h_nav=None, trotter_order=1):

        self.up_then_down = up_then_down
        self.mapping = mapping
        if intervals > 1:
            self.intervals = intervals
        else:
            raise ValueError("The number of intervals must be greater than 1.")
        self.time = time
        self.dt = self.time/self.intervals
        self.reference_state = reference_state
        if trotter_order in [1, 2]:
            self.trotter_order = trotter_order
        else:
            raise ValueError("Only trotter_order = 1, 2 is supported")

        if molecule is None:
            self.qubit_hamiltonian = qubit_hamiltonian
            if not isinstance(h_init, ofQubitOperator):
                raise ValueError("When providing a qubit hamiltonian, an initial qubit Hamiltonian must also be provided")
            self.h_init = h_init
            if not isinstance(reference_state, Circuit):
                raise ValueError("Reference state Circuit must be provided when simulating a qubit hamiltonian directly")
            self.reference_state = reference_state
        else:
            self.n_electrons = molecule.n_active_electrons
            self.n_spinorbitals = int(molecule.n_sos)
            self.n_qubits = get_qubit_number(mapping, self.n_spinorbitals)
            self.spin = molecule.spin
            self.qubit_hamiltonian = fermion_to_qubit_mapping(molecule.fermionic_hamiltonian, n_spinorbitals=self.n_spinorbitals,
                                                              n_electrons=self.n_electrons, mapping=self.mapping,
                                                              up_then_down=self.up_then_down, spin=self.spin)
            self.h_init = self._build_h_init(molecule) if h_init is None else h_init
        self.h_final = self.qubit_hamiltonian
        self.h_nav = h_nav

        def qu_op_to_list(qu_op):
            """Remove consant term and convert QubitOperator to list of (term, coeff)"""
            new_qu_op = qu_op - qu_op.constant
            new_qu_op.compress()
            return list(new_qu_op.terms.items())

        self.h_final_list = qu_op_to_list(self.h_final)
        self.n_h_final = len(self.h_final_list)
        self.h_init_list = qu_op_to_list(self.h_init)
        self.n_h_init = len(self.h_init_list)
        if self.h_nav is None:
            self.stride = 2
            self.n_h_nav = 0
        else:
            if isinstance(self.h_nav, ofQubitOperator):
                self.stride = 3
                self.h_nav_list = qu_op_to_list(self.h_nav)
                self.n_h_nav = len(self.h_nav_list)
            else:
                raise ValueError("Navigator Hamiltonian must be a QubitOperator")

        self.n_var_params = (intervals - 1) * self.stride
        self.n_var_gates = (self.n_h_init + self.n_h_final + self.n_h_nav) * self.trotter_order

        self.var_params = None
        self.circuit = None

    def _build_h_init(self, molecule):
        """Return the initial Hamiltonian (h_init) composed of the one-body terms derived from the diagonal of Fock
        matrix and one-body off-diagonal terms"""
        core_constant, h1, two_body = molecule.get_active_space_integrals()
        diag_fock = np.diag(h1).copy()
        n_active_occupied = len(molecule.active_occupied)
        for j in range(molecule.n_active_mos):
            for i in range(n_active_occupied):
                diag_fock[j] += 2*two_body[i, j, j, i] - 1*two_body[i, j, i, j]

        hf_ferm = FermionOperator((), core_constant)
        for i in range(molecule.n_active_mos):
            for j in range(molecule.n_active_mos):
                if i != j:
                    hf_ferm += FermionOperator(((i * 2, 1), (j * 2, 0)), h1[i, j])
                    hf_ferm += FermionOperator(((i * 2 + 1, 1), (j * 2 + 1, 0)), h1[i, j])
                else:
                    hf_ferm += FermionOperator(((i * 2, 1), (j * 2, 0)), diag_fock[i])
                    hf_ferm += FermionOperator(((i * 2 + 1, 1), (j * 2 + 1, 0)), diag_fock[j])
        return fermion_to_qubit_mapping(hf_ferm, mapping=self.mapping, n_spinorbitals=self.n_spinorbitals, n_electrons=self.n_electrons,
                                        up_then_down=self.up_then_down, spin=self.spin)

    def set_var_params(self, var_params=None):
        """Set values for the variational parameters. Default is linear interpolation."""
        if var_params is None:
            var_params = self._init_params()[self.stride:self.n_var_params+self.stride]

        init_var_params = np.array(var_params)
        if init_var_params.size == self.n_var_params:
            self.var_params = var_params
        else:
            raise ValueError(f"Expected {self.n_var_params} variational parameters but received {init_var_params.size}.")
        return var_params

    def update_var_params(self, var_params):
        """Update the variational parameters in the circuit without rebuilding."""
        for i in range(self.intervals-1):
            self._update_gate_params_for_qu_op(self.h_init_list, self.n_var_gates * i, var_params[self.stride*i], self.n_h_init)
            self._update_gate_params_for_qu_op(self.h_final_list, self.n_var_gates * i + self.n_h_init * self.trotter_order,
                                               var_params[self.stride*i+1], self.n_h_final)
            if self.h_nav is not None:
                self._update_gate_params_for_qu_op(self.h_nav_list, self.n_var_gates * i + (self.n_h_init + self.n_h_final) * self.trotter_order,
                                                   var_params[self.stride*i+2], self.n_h_nav)

    def _update_gate_params_for_qu_op(self, qu_op_list, n_var_start, var_param, num_terms):
        """Updates the corresponding circuit variational_gates for a QubitOperator defined by term order qu_op_list

        Args:
            qu_op_list :: The list with elements (term, coeff) that defines the trotterization of a QubitOperator
            n_var_start :: The varational_gates position that the trotterization of qu_op starts
            var_param :: The variational parameter (evolution time) for qu_op. Same for all terms in qu_op
            num_terms :: The number of terms in qu_op
        """
        prefac = 2 / self.trotter_order * self.dt * var_param
        for i, (_, coeff) in enumerate(qu_op_list):
            self.circuit._variational_gates[n_var_start+i].parameter = prefac * coeff
        if self.trotter_order == 2:
            for i, (_, coeff) in enumerate(list(reversed(qu_op_list))):
                self.circuit._variational_gates[n_var_start+num_terms+i].parameter = prefac * coeff

    def _init_params(self):
        """Generate initial parameters for the VSQS algorithm.
        a_i = step*i, b_i=1-step*i, c_i= 1-step*i i<=n_intervals/2, step*i i>n_intervals/2
        """
        a = np.zeros(self.intervals+1)
        b = np.zeros(self.intervals+1)
        a[0] = 1
        b[self.intervals] = 1
        step_size = 1/self.intervals
        for i in range(1, self.intervals):
            a[i] = (1 - i * step_size)
            b[i] = (i * step_size)
        all_params = np.zeros(self.stride * (self.intervals + 1))
        if self.h_nav is None:
            # order [a[0], b[0], a[1], b[1],...]
            all_params = np.dstack((a, b)).flatten()
        else:
            c = np.zeros(self.intervals+1)
            c[0:self.intervals//2] = b[0:self.intervals//2]
            c[self.intervals//2:self.intervals+1] = a[self.intervals//2:self.intervals+1]
            # order [a[0], b[0], c[0], a[1], b[1], c[1],...]
            all_params = np.dstack((a, b, c)).flatten()
        return all_params

    def prepare_reference_state(self):
        """Prepare a circuit generating the HF reference state."""
        return get_reference_circuit(n_spinorbitals=self.n_spinorbitals, n_electrons=self.n_electrons, mapping=self.mapping,
                                     up_then_down=self.up_then_down, spin=self.spin)

    def build_circuit(self, var_params=None):
        """Build the VSQS circuit by successive first- or second-order trotterizations of h_init, h_final and possibly h_nav"""
        reference_state_circuit = self.prepare_reference_state() if self.reference_state is None else self.reference_state
        self.var_params = self.set_var_params(var_params)

        vsqs_circuit = get_exponentiated_qubit_operator_circuit(self.h_init, time=self.dt, trotter_order=self.trotter_order, pauli_order=self.h_init_list)
        for i in range(self.intervals-1):
            vsqs_circuit += get_exponentiated_qubit_operator_circuit(self.h_init, time=self.var_params[i * self.stride] * self.dt, variational=True,
                                                                     trotter_order=self.trotter_order, pauli_order=self.h_init_list)
            vsqs_circuit += get_exponentiated_qubit_operator_circuit(self.h_final, time=self.var_params[i * self.stride + 1] * self.dt, variational=True,
                                                                     trotter_order=self.trotter_order, pauli_order=self.h_final_list)
            if self.h_nav is not None:
                vsqs_circuit += get_exponentiated_qubit_operator_circuit(self.h_nav, time=self.var_params[i * self.stride + 2] * self.dt, variational=True,
                                                                         trotter_order=self.trotter_order, pauli_order=self.h_nav_list)
        vsqs_circuit += get_exponentiated_qubit_operator_circuit(self.h_final, time=self.dt, trotter_order=self.trotter_order, pauli_order=self.h_final_list)

        self.circuit = reference_state_circuit + vsqs_circuit if reference_state_circuit.size != 0 else vsqs_circuit

        return self.circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import numpy as np
import os
from openfermion import load_operator

from tangelo.molecule_library import mol_H2_sto3g, mol_H4_doublecation_minao, mol_H4_cation_sto3g
from tangelo.toolboxes.qubit_mappings import jordan_wigner
from tangelo.toolboxes.ansatz_generator.upccgsd import UpCCGSD
from tangelo.linq import Simulator

# For openfermion.load_operator function.
pwd_this_test = os.path.dirname(os.path.abspath(__file__))


class UpCCGSDTest(unittest.TestCase):

    def test_upccgsd_set_var_params(self):
        """Verify behavior of set_var_params for different inputs (keyword,
        list, numpy array).
        """

        upccgsd_ansatz = UpCCGSD(mol_H2_sto3g)

        six_ones = np.ones((6,))

        upccgsd_ansatz.set_var_params("ones")
        np.testing.assert_array_almost_equal(upccgsd_ansatz.var_params, six_ones, decimal=6)

        upccgsd_ansatz.set_var_params([1., 1., 1., 1., 1., 1.])
        np.testing.assert_array_almost_equal(upccgsd_ansatz.var_params, six_ones, decimal=6)

        upccgsd_ansatz.set_var_params(np.array([1., 1., 1., 1., 1., 1.]))
        np.testing.assert_array_almost_equal(upccgsd_ansatz.var_params, six_ones, decimal=6)

    def test_upccgsd_incorrect_number_var_params(self):
        """Return an error if user provide incorrect number of variational
        parameters.
        """

        upccgsd_ansatz = UpCCGSD(mol_H2_sto3g)

        self.assertRaises(ValueError, upccgsd_ansatz.set_var_params, np.array([1., 1., 1., 1.]))

    def test_upccgsd_H2(self):
        """Verify closed-shell UpCCGSD functionalities for H2."""

        # Build circuit
        upccgsd_ansatz = UpCCGSD(mol_H2_sto3g)
        upccgsd_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = jordan_wigner(mol_H2_sto3g.fermionic_hamiltonian)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        upccgsd_ansatz.update_var_params([0.03518165, -0.02986551,  0.02897598, -0.03632711,
                                          0.03044071,  0.08252277])
        energy = sim.get_expectation_value(qubit_hamiltonian, upccgsd_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.1372658, delta=1e-6)

    def test_upccgsd_H4_open(self):
        """Verify open-shell UpCCGSD functionalities for H4."""

        var_params = [-0.0291763,   0.36927821,  0.14654907, -0.13845063,  0.14387348, -0.00903457,
                      -0.56843484,  0.01223853,  0.13649942,  0.83225887,  0.20236275,  0.02682977,
                      -0.17198068,  0.10161518,  0.01523924,  0.30848876,  0.22430705, -0.07290468,
                       0.16253591,  0.02268874,  0.2382988,   0.33716289, -0.20094664,  0.3057071,
                      -0.58426117,  0.22433297,  0.29668267,  0.64761217, -0.2705204,   0.07540534,
                       0.20131878,  0.09890588,  0.10563459, -0.22983007,  0.13578206, -0.02017009]

        # Build circuit
        upccgsd_ansatz = UpCCGSD(mol_H4_cation_sto3g)
        upccgsd_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = jordan_wigner(mol_H4_cation_sto3g.fermionic_hamiltonian)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        upccgsd_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, upccgsd_ansatz.circuit)
        self.assertAlmostEqual(energy, -1.6412047312, delta=1e-6)

    def test_upccgsd_H4_doublecation(self):
        """Verify closed-shell UpCCGSD functionalities for H4 2+."""

        var_params = [1.08956248, 1.08956247, 1.05305993, 1.05305993, 0.8799399, 0.8799399,
                      0.88616586, 0.88616586, 1.09532143, 1.09532143, 1.23586857, 1.23586857,
                      1.09001216, 0.85772769, 1.28020861, 1.05820721, 0.9680792,  1.01693601,
                      0.68355852, 0.68355852, 1.30303827, 1.30303827, 0.74524063, 0.74524063,
                      0.36958813, 0.36958813, 1.37092805, 1.37092805, 0.92860293, 0.92860293,
                      1.30296676, 0.5803438,  1.42469953, 1.05666723, 0.86961358, 0.55347531]

        # Build circuit
        upccgsd_ansatz = UpCCGSD(mol_H4_doublecation_minao)
        upccgsd_ansatz.build_circuit()

        # Build qubit hamiltonian for energy evaluation
        qubit_hamiltonian = load_operator("mol_H4_doublecation_minao_qubitham_jw.data", data_directory=pwd_this_test+"/data", plain_text=True)

        # Assert energy returned is as expected for given parameters
        sim = Simulator()
        upccgsd_ansatz.update_var_params(var_params)
        energy = sim.get_expectation_value(qubit_hamiltonian, upccgsd_ansatz.circuit)
        self.assertAlmostEqual(energy, -0.854608, delta=1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from openfermion.transforms.opconversions.conversions import get_majorana_operator

from tangelo.toolboxes.operators import QubitOperator

sigma_map = {"0": "X", "1": "Y", "2": "Z"}


def _node_value(p, l):
    """Obtain the node value from vector p and depth l as given in Eq (3) in
    arXiv:1910.10746v2"""
    nv = (3**l - 1)//2
    for j in range(l):
        nv += 3**(l-1-j)*int(p[j])
    return nv


def _jkmn_list(h):
    """Generate the full list of paths in the JKMN defined ternary tree of height h

    The algorithm is described in arXiv:1910.10746v2

    Args:
        h (int) : The height of the full ternary tree to generate the leaf paths

    Returns:
        list : The list with elements that generate the qubit operators
        for each path in the ternary tree
    """
    t_list = list()
    for i in range(3**h):
        # string representation of leaf index with base 3 and padding 0s to the tree height h
        terstring = np.base_repr(i, base=3).rjust(h, '0')

        # descend ternary tree and obtain tuples describing QubitOperator
        c_qu_op = []
        for ch in range(h):
            # generates tuple that initiates QubitOperator on index nv and operator op=X,Y,Z
            nv = _node_value(terstring, ch)
            op = sigma_map[terstring[ch]]
            c_qu_op += [(nv, op)]
        t_list += [c_qu_op]
    return t_list


def _jkmn_dict(n_qubits):
    r"""Generate the mapping from Majorana modes to qubit operators

    The algorithm is described in arXiv:1910.10746v2

    Although all assignments of leaves to Majorana modes are valid (with one arbitrary leaf being dropped),
    this function assigns leaves to majorana modes via the following process
    1) The furthest right leaf is the mode dropped.
    2) One can obtain the vaccuum state by applying Hadamard transforms to certain qubits.
       This essentially reorders the assignment of leafs to Majorana modes as HXH=Z, HZH=X, HYH=-Y, HH=I
    3) The order is finalized by assigning \gamma_{2*i} with 'X' on qubit i and \gamma_{2*i+1}
       has 'Y' on qubit i. If \gamma_{2*i} and \gamma_{2*i+1} were swapped, a negative sign on one of the
       \gamma_{2*i} or \gamma_{2*i+1} Majorana modes would be required to obtain a valid vaccuum state.

    This process ensures that all number operators a_i^{\dagger}a_i = 1/2*(1 + 1j*\gamma_{2*i}\gamma_{2*j+1})
    result in qubit operators with only Z operators including Z on qubit i.

    Args:
        n_qubits (int) : The number of qubits in the system

    Returns:
        dict : The mapping from Majorana modes to qubit operators
    """

    # Calculate initial height of ternary tree (largest full tree with n < n_qubit nodes)
    h = int(np.log10(2 * n_qubits + 1) / np.log10(3))

    # create full tree of largest size with less than n nodes
    all_list = _jkmn_list(h)

    # Obtain number of leaves of tree that need to be branched
    n_leaves_to_qubit = n_qubits - (3**h - 1) // 2

    # change first n_leaves_to_qubit leaves to qubits
    prepend_list = []
    for j in range(n_leaves_to_qubit):
        # remove leaf and change to node append sigma_x sigma_y sigma_z
        item_to_qubit = all_list.pop(0)
        for i in range(3):
            nv = (3**(h)-1)//2 + j
            prepend_list.append(item_to_qubit+[(nv, sigma_map[str(i)])])

    all_list = prepend_list + all_list

    # Create map from Majorana modes to QubitOperators ignoring furthest right leaf
    jkmn_map = dict()
    for i, v in enumerate(all_list[:-1]):
        jkmn_map[i] = QubitOperator(v)

    # Find indices to apply H transformation HXH=Z, HZH=X, HYH=-Y
    hinds = _jkmn_vaccuum_indices(n_qubits, jkmn_map)

    # Obtain transformed jkmn map and determine \gamma_{2*i} \gamma_{2*i+1} majorana modes to swap
    tjkmn_map = dict()
    for i in range(2*n_qubits):
        t = tuple()
        for term in jkmn_map[i].terms.keys():
            for iterm in term:
                if iterm[0] in hinds:
                    if iterm[1] == 'X':
                        t += ((iterm[0], 'Z'),)
                    elif iterm[1] == 'Z':
                        t += ((iterm[0], 'X'),)
                    else:
                        t += ((iterm[0], iterm[1]), )
                else:
                    t += ((iterm[0], iterm[1]),)
        tjkmn_map[i] = QubitOperator(t)

    # Swap order of majorana modes so that \gamma_{2*i} has 'X' on qubit i and \gamma_{2*i+1} has 'Y' on qubit i
    stjkmn_map = dict()
    for i in range(n_qubits):
        q1 = next(iter(tjkmn_map[2*i].terms))
        q2 = next(iter(tjkmn_map[2*i+1].terms))
        for tup in q1:
            if tup[1] == 'X' and (tup[0], 'Y') in q2:
                stjkmn_map[2*tup[0]] = tjkmn_map[2*i]
                stjkmn_map[2*tup[0]+1] = tjkmn_map[2*i+1]
            elif tup[1] == 'Y' and (tup[0], 'X') in q2:
                stjkmn_map[2*tup[0]+1] = tjkmn_map[2*i]
                stjkmn_map[2*tup[0]] = tjkmn_map[2*i+1]

    return stjkmn_map


def jkmn(fermion_operator, n_qubits):
    """The JKMN mapping of a fermion operator as described in arXiv:1910.10746v2

    Args:
        fermion_operator (FermionOperator) : The fermion operator to transform to a qubit operator
        n_qubits (int) : The number of qubits in the system

    Returns:
        QubitOperator : The qubit operator corresponding to the Fermion Operator
    """
    mj_op = get_majorana_operator(fermion_operator)
    jkmn_map = _jkmn_dict(n_qubits)

    full_operator = QubitOperator()
    for term, coeff in mj_op.terms.items():
        if abs(coeff) > 1.e-12:
            c_term = QubitOperator([], coeff)
            for mj in term:
                c_term *= jkmn_map[mj]
            full_operator += c_term
    full_operator.compress()
    return full_operator


def _jkmn_vaccuum_indices(n_qubits, jkmn_map=None):
    """Generate the indices for which an 'H' transformation is necessary

    Args:
        n_qubits (int) : The number of qubits in the system
        jkmn_map (dict) : The mapping from a Majorana operator index to a QubitOperator

    Returns:
        list : Indices that require an 'H' gate transformation
    """

    if jkmn_map is None:
        jkmn_map = _jkmn_dict(n_qubits)
    fock_dict = dict()
    rot_dict = dict()
    # generate number operators
    for i in range(n_qubits):
        fock_dict[i] = 0.5 * QubitOperator([]) - 0.5j * jkmn_map[2*i] * jkmn_map[2*i+1]
        for k in fock_dict[i].terms:
            for i in k:
                if i[0] not in rot_dict:
                    rot_dict[i[0]] = i[1]

    return [int(k) for k, v in rot_dict.items() if v == "X"]


def jkmn_prep_vector(vector):
    r"""Apply JKMN mapping to fermion occupation vector.

    Each fermionic mode i is generated by applying \gamma_{2*i} = a_i^{\dagger} + a_i. The returned
    vector is defined by which qubits have X or Y operations applied.

    Args:
        vector (list of int) : The occupation of each spinorbital

    Returns:
        list[int] : The state preparation vector that defines which qubits to apply X gates to
    """
    n_qubits = len(vector)
    jkmn_map = _jkmn_dict(n_qubits)

    state_prep_qu_op = QubitOperator([], 1)
    for i, occ in enumerate(vector):
        if occ == 1:
            state_prep_qu_op *= jkmn_map[2*i]

    x_vector = np.zeros(n_qubits, dtype=int)
    for k in state_prep_qu_op.terms:
        for i in k:
            x_vector[i[0]] = 1 if i[1] in ["X", "Y"] else 0
    return list(x_vector)



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module to create Hardware Efficient Ansatze (HEA) circuit with n_layers."""

from tangelo.linq import Circuit, Gate


def rotation_circuit(n_qubits, rot_type="euler"):
    """Construct a circuit applying an Euler Z-X-Z rotation to each qubit."""
    if rot_type == "euler":
        gateslist = [Gate("RZ", target=target, is_variational=True) for target in range(n_qubits)]
        gateslist.extend(Gate("RX", target=target, is_variational=True) for target in range(n_qubits))
        gateslist.extend(Gate("RZ", target=target, is_variational=True) for target in range(n_qubits))
    elif rot_type == "real":
        gateslist = [Gate("RY", target, is_variational=True) for target in range(n_qubits)]
    else:
        raise ValueError(f"Supported keywords for initializing variational parameters: ['real', 'euler'']")
    return Circuit(gateslist)


def entangler_circuit(n_qubits):
    """Construct a circuit applying two columns of staggered CNOT gates to all
    qubits and their neighbours.
    """

    gateslist = [Gate("CNOT", control=2*ii, target=2*ii + 1) for ii in range(n_qubits//2)]
    gateslist.extend(Gate("CNOT", control=2*ii + 1, target=2*(ii+1)) for ii in range(n_qubits//2 - 1))

    return Circuit(gateslist)


def construct_hea_circuit(n_qubits, n_layers, rot_type="euler"):
    """Construct a circuit consisting of alternating sequence of Euler rotations
    and entanglers.
    """
    circuit = rotation_circuit(n_qubits, rot_type)
    for ii in range(n_layers):
        circuit += entangler_circuit(n_qubits)
        circuit += rotation_circuit(n_qubits, rot_type)
    return circuit



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module to regroup the list of possible atoms and built-in chemical groups for
broken link capping. The geometries for the capping groups are taken from the
NIST Chemistry WebBook database (https://webbook.nist.gov/chemistry/).
"""


elements = [
    "X",
    "H", "He", "Li", "Be",  "B",  "C",  "N",  "O",  "F", "Ne",
    "Na", "Mg", "Al", "Si",  "P",  "S", "Cl", "Ar",  "K", "Ca",
    "Sc", "Ti",  "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn",
    "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr",  "Y", "Zr",
    "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn",
    "Sb", "Te",  "I", "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd",
    "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb",
    "Lu", "Hf", "Ta",  "W", "Re", "Os", "Ir", "Pt", "Au", "Hg",
    "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th",
    "Pa",  "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm",
    "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds",
    "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts", "Og"
]

chemical_groups = {
    "CH3": [  # From ethane.
        ["X", [  2.15410,  0.87450,  1.16590]],
        ["C", [  0.71660,  0.89800,  0.64250]],
        ["H", [  0.53970,  1.76660, -0.00250]],
        ["H", [  0.48990,  0.00050,  0.05510]],
        ["H", [ -0.00780,  0.94520,  1.46400]]
    ],
    "CF3": [  # From hexafluoroethane.
        ["X", [  2.03950,  1.01440,  0.00010]],
        ["C", [  0.49550,  1.01460, -0.00010]],
        ["F", [  0.04100, -0.24410,  0.00000]],
        ["F", [  0.04120,  1.64380, -1.09040]],
        ["F", [  0.04080,  1.64420,  1.08980]]
    ],
    "NH2": [  # From ammonia.
        ["X", [ -0.01300, -0.00270,  0.00580]],
        ["N", [  0.87140,  0.50430, -0.01000]],
        ["H", [  1.59990, -0.20890,  0.00580]],
        ["H", [  0.92660,  0.93550, -0.93240]]
    ]
}



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions helping with quantum circuit format conversion between abstract
format and ionq format.

In order to produce an equivalent circuit for the target backend, it is
necessary to account for:
- how the gate names differ between the source backend to the target backend.
- how the order and conventions for some of the inputs to the gate operations
    may also differ.
"""
from math import pi, isclose


def get_ionq_gates():
    """Map gate name of the abstract format to the equivalent gate name used in
    the json IonQ format. For more information:
    - https://dewdrop.ionq.co/
    - https://docs.ionq.co
    """

    GATE_JSON_IONQ = dict()
    for name in {"H", "X", "Y", "Z", "S", "T", "RX", "RY", "RZ", "CNOT", "PHASE"}:
        GATE_JSON_IONQ[name] = name.lower()
    return GATE_JSON_IONQ


def translate_json_ionq(source_circuit):
    """Take in an abstract circuit, return a dictionary following the IonQ JSON
    format as described below.
    https://dewdrop.ionq.co/#json-specification

    Args:
        source_circuit: quantum circuit in the abstract format.

    Returns:
        dict: representation of the quantum circuit following the IonQ JSON
            format.
    """

    GATE_JSON_IONQ = get_ionq_gates()

    json_gates = []
    for gate in source_circuit._gates:
        if gate.name in {"H", "X", "Y", "Z", "S", "T"}:
            json_gates.append({'gate': GATE_JSON_IONQ[gate.name], 'target': gate.target[0]})
        elif gate.name in {"RX", "RY", "RZ"}:
            json_gates.append({'gate': GATE_JSON_IONQ[gate.name], 'target': gate.target[0], 'rotation': gate.parameter})
        elif gate.name in {"PHASE"}:
            if isclose(gate.parameter, pi / 2, abs_tol=1.e-7):
                json_gates.append({'gate': 's', 'target': gate.target[0]})
            elif isclose(gate.parameter, - pi / 2, abs_tol=1.e-7):
                json_gates.append({'gate': 'si', 'target': gate.target[0]})
            elif isclose(gate.parameter, pi / 4, abs_tol=1.e-7):
                json_gates.append({'gate': 't', 'target': gate.target[0]})
            elif isclose(gate.parameter, - pi / 4, abs_tol=1.e-7):
                json_gates.append({'gate': 'ti', 'target': gate.target[0]})
            else:
                raise ValueError(f"Only phases of pi/2, -pi/2, pi/4, -pi/4 are supported with JSON IonQ translation")
        elif gate.name in {"CNOT"}:
            json_gates.append({'gate': GATE_JSON_IONQ[gate.name], 'target': gate.target[0], 'control': gate.control[0]})
        else:
            raise ValueError(f"Gate '{gate.name}' not supported with JSON IonQ translation")

    json_ionq_circ = {"qubits": source_circuit.width, 'circuit': json_gates}
    return json_ionq_circ



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the fermionic operators that can be used to generate
the UpCCGSD ansatz.
"""

from tangelo.toolboxes.ansatz_generator._general_unitary_cc import get_spin_ordered
from tangelo.toolboxes.operators import FermionOperator


def get_upccgsd(n_orbs, values, up_then_down=False, anti_hermitian=True):
    r"""Function to generate one UpCCGSD layer as a FermionicOperator

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2)
        values (list): parameters for 1-UpCCGSD
        up_then_down (bool): The ordering of the spin orbitals. qiskit (True) openfermion (False)
        anit_hermitian (bool): Whether to include the anti_hermitian conjugate in the operator

    Returns:
        ferm_op (FermionicOperator): The 1-UpCCGSD ansatz as a fermionic operator"""

    all_terms = get_generalized_singles(n_orbs, up_then_down) + get_paired_doubles(n_orbs, up_then_down)

    ferm_op = FermionOperator()
    for i, item in enumerate(all_terms):
        ferm_op += FermionOperator(item[0], values[i])
        if anti_hermitian:
            if len(item[0]) == 2:
                ferm_op += FermionOperator(((item[0][1][0], 1), (item[0][0][0], 0)), -values[i])
            else:   # length is 4, i.e. generalized paired double term
                ferm_op += FermionOperator(((item[0][3][0], 1), (item[0][2][0], 0),
                                            (item[0][1][0], 1), (item[0][0][0], 0)), -values[i])
    return ferm_op


def get_generalized_singles(n_orbs, up_then_down=False):
    r"""Function to obtain the generalized singles fermionic generator

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2)
        up_then_down (bool): The ordering of the spin orbitals. qiskit (True) openfermion (False)

    Returns:
        all_terms (list): The generalized singles term \sum_{ij} t_{ij} a_i^{\dagger} a_j"""

    all_terms = list()
    for i in range(n_orbs):
        for j in range(i+1, n_orbs):
            up, dn = get_spin_ordered(n_orbs, j, i, up_down=up_then_down)  # get spin-orbital indices
            all_terms.extend([[((up[0], 1), (up[1], 0)), 1], [((dn[0], 1), (dn[1], 0)), 1]])
    return all_terms


def get_paired_doubles(n_orbs, up_then_down=False):
    r"""Function to obtain the generalized paired doubles fermionic generator

    Args:
        n_orbs (int): number of orbitals in the fermion basis (this is number of
            spin-orbitals divided by 2)
        up_then_down (bool): The ordering of the spin orbitals. qiskit (True) openfermion (False)

    Returns:
        all_terms (list): The paired doubles term \sum_{ij} t_{ij} a_i^+ a_j a_i^+ a_j"""

    all_terms = list()
    for i in range(n_orbs):
        for j in range(i+1, n_orbs):
            up, dn = get_spin_ordered(n_orbs, j, i, up_down=up_then_down)  # get spin-orbital indices
            all_terms.extend([[((up[0], 1), (up[1], 0), (dn[0], 1), (dn[1], 0)), 1]])
    return all_terms



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements the iterative qubit coupled cluster (iQCC)-VQE
procedure of Ref. 1. It is a variational approach that utilizes the
the QCC ansatz to produce shallow circuits. The iterative procedure
allows a small number (1—10) of generators to be used for the QCC
This results in even shallower circuits and fewer quantum resources
for the iQCC approach relative to the native QCC method. A caveat
is that after each iteration, the qubit Hamiltonian is dressed with
the generators and optimal parameters, the result of which is an
exponential growth of the number of terms. A technique also described
in Ref. 1 can be utilized to address this issue by discarding some
terms based on the Frobenius norm of the Hamiltonian.

Refs:
    1. I. G. Ryabinkin, R. A. Lang, S. N. Genin, and A. F. Izmaylov.
        J. Chem. Theory Comput. 2020, 16, 2, 1055–1063.
"""

from tangelo.linq import Simulator
from tangelo.toolboxes.ansatz_generator.qcc import QCC
from tangelo.algorithms.variational.vqe_solver import VQESolver
from tangelo.toolboxes.ansatz_generator._qubit_cc import qcc_op_dress


class iQCC_solver:
    """The iQCC-VQE solver class combines the QCC ansatz and VQESolver classes
    to perform an iterative and variational procedure to compute the total QCC
    energy for a given Hamiltonian. The algorithm is outlined below:

    (0) Prepare a qubit Hamiltonian, initialize QMF parameters, construct the
        DIS, select QCC generators, and initialize QCC amplitudes.
    (1) Simulate the QCC energy through VQE minimization.
    (2) Check if the energy is lowered relative to the previous iteration.
    (3) If the energy is lowered, proceed to (4); else, keep the QCC generators,
        re-initialize the amplitudes, and re-compute the energy. If after several
        attempts the energy is not lowered, set all QCC amplitudes to zero and
        use the QMF parameters from the previous iteration to compute the energy.
        This is guaranteed to yield a lower energy.
    (4) Check termination criteria: terminate if the change in energy is below a
        threshold, the DIS is empty, or the maximum number of iterations is reached.
    (5) If not terminated, dress the qubit Hamiltonian with the current QCC
        generators and optimal amplitudes.
    (6) Purify the QMF parameters, rebuild the DIS, and select generators for
        the next iteration; return to (1) and repeat until termination.

    Attributes:
        molecule (SecondQuantizedMolecule): The molecular system.
        qubit_mapping (str): One of the supported qubit mapping identifiers. Default, "jw".
        up_then_down (bool): Change basis ordering putting all spin up orbitals first,
            followed by all spin down. Default, False.
        initial_var_params (str or array-like): Initial values of the variational parameters
            for the classical optimizer.
        backend_options (dict): Parameters to build the tangelo.linq Simulator
            class.
        penalty_terms (dict): Parameters for penalty terms to append to target
            qubit Hamiltonian (see penaly_terms for more details).
        ansatz_options (dict): Parameters for the chosen ansatz (see given ansatz
            file for details).
        qubit_hamiltonian (QubitOperator-like): Self-explanatory.
        deqcc_thresh (float): threshold for the difference in iQCC energies between
            consecutive iterations required for convergence of the algorithm.
            Default, 1e-5 Hartree.
        max_iqcc_iter (int): maximum number of iQCC iterations allowed before termination.
            Default, 100.
        max_iqcc_retries (int): if the iQCC energy for a given iteration is not lower than
            the value from the previous iteration, the iQCC parameters are reinitialized
            and the VQE procedure will be attempted up to max_iqcc_retries times. If unsuccessful
            after max_iqcc_retries attempts, the iQCC parameters are all set to 0 and the QMF
            Bloch angles from the previous iteration are used. Default, 10.
        compress_qubit_ham (bool): controls whether the qubit Hamiltonian is compressed
            after dressing with the current set of generators at the end of each iQCC iteration.
            Default, False.
        compress_eps (float): parameter required for compressing intermediate iQCC Hamiltonians
            using the Froebenius norm. Discarding terms in this manner will not alter the
            eigenspeectrum of intermediate Hamiltonians by more than compress_eps.
            Default, 1.59e-3 Hartree.
        verbose (bool): Flag for verbosity of iQCCsolver. Default, False.
     """

    def __init__(self, opt_dict):

        default_backend_options = {"target": None, "n_shots": None, "noise_model": None}
        default_options = {"molecule": None,
                           "qubit_mapping": "jw",
                           "up_then_down": False,
                           "initial_var_params": None,
                           "backend_options": default_backend_options,
                           "penalty_terms": None,
                           "ansatz_options": dict(),
                           "qubit_hamiltonian": None,
                           "deqcc_thresh": 1e-5,
                           "max_iqcc_iter": 100,
                           "max_iqcc_retries": 10,
                           "compress_qubit_ham": False,
                           "compress_eps": 1.59e-3,
                           "verbose": False}

        # Initialize with default values
        self.__dict__ = default_options
        # Overwrite default values with user-provided ones, if they correspond to a valid keyword
        for param, val in opt_dict.items():
            if param in default_options:
                setattr(self, param, val)
            else:
                raise KeyError(f"Keyword :: {param}, not available in iQCCsolver")

        if not self.molecule:
            raise ValueError("An instance of SecondQuantizedMolecule is required for initializing iQCCsolver.")

        # initialize variables and lists to store useful data from each iQCC-VQE iteration
        self.energies = []
        self.iteration = 0
        self.converged = False
        self.qmf_energy = None
        self.qcc_ansatz = None
        self.vqe_solver = None
        self.vqe_solver_options = None
        self.final_optimal_energy = None
        self.final_optimal_qmf_params = None
        self.final_optimal_qcc_params = None

    def build(self):
        """Builds the underlying objects required to run the iQCC-VQE algorithm."""

        # instantiate the QCC ansatz but do not build it here because vqe_solver builds it
        self.qcc_ansatz = QCC(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)

        # build an instance of VQESolver with options that remain fixed during the iQCC-VQE routine
        self.vqe_solver_options = {"molecule": self.molecule,
                                   "qubit_mapping": self.qubit_mapping,
                                   "ansatz": self.qcc_ansatz,
                                   "initial_var_params": self.initial_var_params,
                                   "backend_options": self.backend_options,
                                   "penalty_terms": self.penalty_terms,
                                   "up_then_down": self.up_then_down,
                                   "qubit_hamiltonian": self.qubit_hamiltonian,
                                   "verbose": self.verbose}
        self.vqe_solver = VQESolver(self.vqe_solver_options)
        self.vqe_solver.build()

    def simulate(self):
        """Executes the iQCC-VQE algorithm. During each iteration,
        QCC-VQE minimization is performed."""

        # initialize quantities; compute the QMF energy and set this as eqcc_old
        sim = Simulator()
        self.qmf_energy = sim.get_expectation_value(self.qcc_ansatz.qubit_ham, self.qcc_ansatz.qmf_circuit)
        e_qcc, eqcc_old, delta_eqcc = 0., self.qmf_energy, self.deqcc_thresh

        if self.verbose:
            print(f"The qubit mean field energy = {self.qmf_energy}")

        while not self.converged and self.iteration < self.max_iqcc_iter:
            # check that the DIS has at least one generator to use; otherwise terminate
            if self.qcc_ansatz.dis and self.qcc_ansatz.var_params.any():
                e_qcc = self.vqe_solver.simulate()
                delta_eqcc = e_qcc - eqcc_old
                eqcc_old = e_qcc
            else:
                self.converged = True
                if self.verbose:
                    print("Terminating the iQCC-VQE solver: the DIS of QCC generators is empty.")

            # check if unsuccessful: energy is not lowered and energy is not converged.
            if delta_eqcc > 0. and delta_eqcc >= self.deqcc_thresh:
                n_retry = 0
                if self.verbose:
                    print(f"The energy at iteration {self.iteration} is greater than the energy "
                          f"from the previous iteration. Making {self.max_iqcc_retries} attempts "
                          f"to find a lower energy solution")

                # make several attempts to obtain a lower energy
                while e_qcc > eqcc_old and n_retry < self.max_iqcc_retries:
                    self.qcc_ansatz.var_params = None
                    self.qcc_ansatz.update_var_params("random")
                    self.vqe_solver.initial_var_params = self.qcc_ansatz.var_params
                    e_qcc = self.vqe_solver.simulate()
                    n_retry += 1

                # check if energy was lowered; else zero the amplitudes and recompute
                if e_qcc < eqcc_old:
                    delta_eqcc = e_qcc - eqcc_old
                    eqcc_old = e_qcc
                else:
                    self.qcc_ansatz.var_params = None
                    self.qcc_ansatz.update_var_params("qmf_state")
                    self.vqe_solver.initial_var_params = self.qcc_ansatz.var_params
                    eqcc_old = e_qcc
                    e_qcc = self.vqe_solver.simulate()
                    delta_eqcc = e_qcc - eqcc_old

            # update simulation data and check convergence
            if not self.converged:
                self._update_iqcc_solver(delta_eqcc)

        return self.energies[-1]

    def get_resources(self):
        """Returns the quantum resource estimates for the final
        iQCC-VQE iteration."""

        return self.vqe_solver.get_resources()

    def _update_iqcc_solver(self, delta_eqcc):
        """This function serves several purposes after successful iQCC-VQE
        iterations:
            (1) updates/stores the energy, generators, QMF Bloch angles,
                QCC amplitudes, circuits, number of qubit Hamiltonian terms,
                and quantum resource estimates;
            (2) dresses/compresses the qubit Hamiltonian with the current
                generators and optimal amplitudes;
            (3) prepares for the next iteration by rebuilding the DIS,
                re-initializing the amplitudes for a new set of generators,
                generating the circuit, and updates the classical optimizer.
        """

        # get the optimal variational parameters and split them for qmf and qcc
        n_qubits = self.qcc_ansatz.n_qubits
        optimal_qmf_var_params = self.vqe_solver.optimal_var_params[:2*n_qubits]
        optimal_qcc_var_params = self.vqe_solver.optimal_var_params[2*n_qubits:]

        # update all lists with data from the current iteration
        self.energies.append(self.vqe_solver.optimal_energy)

        # dress and (optionally) compress the qubit Hamiltonian
        self.qcc_ansatz.qubit_ham = qcc_op_dress(self.qcc_ansatz.qubit_ham, self.qcc_ansatz.dis,
                                                 optimal_qcc_var_params)
        if self.compress_qubit_ham:
            self.qcc_ansatz.qubit_ham.frobenius_norm_compression(self.compress_eps, n_qubits)

        # set dis and var_params to none to rebuild the dis and initialize new amplitudes
        self.qcc_ansatz.dis = None
        self.qcc_ansatz.var_params = None
        self.qcc_ansatz.build_circuit()
        self.vqe_solver.initial_var_params = self.qcc_ansatz.var_params

        self.iteration += 1

        if self.verbose:
            print(f"Iteration # {self.iteration}")
            print(f"iQCC total energy = {self.vqe_solver.optimal_energy} Eh")
            print(f"iQCC correlation energy = {self.vqe_solver.optimal_energy-self.qmf_energy} Eh")
            print(f"Optimal QMF variational parameters = {optimal_qmf_var_params}")
            print(f"Optimal QCC variational parameters = {optimal_qcc_var_params}")
            print(f"Number of iQCC generators = {len(self.qcc_ansatz.dis)}")
            print(f"iQCC generators = {self.qcc_ansatz.dis}")
            print(f"iQCC resource estimates = {self.get_resources()}")

        if abs(delta_eqcc) < self.deqcc_thresh or self.iteration == self.max_iqcc_iter:
            self.converged = True
            self.final_optimal_energy = self.vqe_solver.optimal_energy
            self.final_optimal_qmf_params = optimal_qmf_var_params
            self.final_optimal_qcc_params = optimal_qcc_var_params

            if self.verbose:
                if abs(delta_eqcc) < self.deqcc_thresh:
                    print("Terminating the iQCC-VQE solver: energy convergence threshold achieved.")
                elif self.iteration == self.max_iqcc_iter:
                    print("Terminating the iQCC-VQE solver: maximum number of iQCC iterations reached.")



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the state-averaged variational quantum eigensolver. Also known as the subspace-search variational quantum
eigensolver.

Ref:
[1] Saad Yalouz, Bruno Senjean, Jakob Gunther, Francesco Buda, Thomas E. O'Brien, Lucas Visscher, "A state-averaged
orbital-optimized hybrid quantum-classical algorithm for a democratic description of ground and excited states",
2021, Quantum Sci. Technol. 6 024004
[2] Ken M Nakanishi, Kosuke Mitarai, Keisuke Fujii, "Subspace-search variational quantum eigensolver for excited states",
Phys. Rev. Research 1, 033062 (2019)
"""

import numpy as np

from tangelo.linq import Simulator, Circuit
from tangelo.toolboxes.operators import qubitop_to_qubitham
from tangelo.toolboxes.qubit_mappings import statevector_mapping
from tangelo.toolboxes.qubit_mappings.mapping_transform import fermion_to_qubit_mapping
from tangelo.toolboxes.ansatz_generator.ansatz import Ansatz
from tangelo.toolboxes.ansatz_generator import UCCSD, HEA, UpCCGSD, UCCGD, VariationalCircuitAnsatz
from tangelo.toolboxes.ansatz_generator.penalty_terms import combined_penalty
from tangelo.algorithms.variational import BuiltInAnsatze, VQESolver


class SA_VQESolver(VQESolver):
    r"""Solve the electronic structure problem for a molecular system by using
    the state-averaged variational quantum eigensolver (SA-VQE) algorithm.

    This algorithm evaluates the energy of a molecular system by performing
    classical optimization over a parametrized ansatz circuit for multiple reference states.

    Users must first set the desired options of the SA_VQESolver object through the
    __init__ method, and call the "build" method to build the underlying objects
    (mean-field, hardware backend, ansatz...). They are then able to call any of
    the energy_estimation, simulate, or get_rdm methods. In particular, simulate
    runs the VQE algorithm, returning the optimal energy found by the classical
    optimizer.

    Attributes:
        molecule (SecondQuantizedMolecule) : the molecular system.
        qubit_mapping (str) : one of the supported qubit mapping identifiers.
        ansatz (Ansatze) : one of the supported ansatze.
        optimizer (function handle): a function defining the classical optimizer and its behavior.
        initial_var_params (str or array-like) : initial value for the classical optimizer.
        backend_options (dict) : parameters to build the tangelo.linq Simulator class.
        penalty_terms (dict): parameters for penalty terms to append to target qubit Hamiltonian (see penalty_terms
            for more details).
        deflation_circuits (list[Circuit]): Deflation circuits to add an orthogonalization penalty with.
        deflation_coeff (float): The coefficient of the deflation.
        ansatz_options (dict): parameters for the given ansatz (see given ansatz file for details).
        up_then_down (bool): change basis ordering putting all spin up orbitals first, followed by all spin down.
            Default, False has alternating spin up/down ordering.
        qubit_hamiltonian (QubitOperator-like): Self-explanatory.
        verbose (bool): Flag for VQE verbosity.
        ref_states (list): The vector occupations of the reference configurations or the reference circuits.
        weights (array): The weights of the occupations
    """

    def __init__(self, opt_dict):

        sa_vqe_options = {"ref_states": None, "weights": None, "ansatz": BuiltInAnsatze.UCCGD}

        # remove SA-VQE specific options before calling VQESolver.__init__() and generate new sa_vqe_options
        opt_dict_vqe = opt_dict.copy()
        for k, v in opt_dict.items():
            if k in sa_vqe_options:
                sa_vqe_options[k] = opt_dict_vqe.pop(k)

        # Initialization of VQESolver will check if spurious dictionary items are present
        super().__init__(opt_dict_vqe)

        self.builtin_ansatze = set([BuiltInAnsatze.UpCCGSD, BuiltInAnsatze.UCCGD, BuiltInAnsatze.HEA, BuiltInAnsatze.UCCSD])

        # Add sa_vqe_options to attributes
        for k, v in sa_vqe_options.items():
            setattr(self, k, v)

        if self.ref_states is None:
            raise ValueError(f"ref_states must be provided when instantiating {self.__class__.__name__}")

        self.n_states = len(self.ref_states)
        if self.weights is None:
            self.weights = np.ones(self.n_states)/self.n_states
        else:
            if len(self.weights) != self.n_states:
                raise ValueError("Number of elements in weights must equal the number of ref_states")
            self.weights = np.array(self.weights)
            self.weights = self.weights/sum(self.weights)

    def build(self):
        """Build the underlying objects required to run the SA-VQE algorithm
        afterwards.
        """

        if isinstance(self.ansatz, Circuit):
            self.ansatz = VariationalCircuitAnsatz(self.ansatz)

        # Building VQE with a molecule as input.
        if self.molecule:

            # Compute qubit hamiltonian for the input molecular system
            qubit_op = fermion_to_qubit_mapping(fermion_operator=self.molecule.fermionic_hamiltonian,
                                                mapping=self.qubit_mapping,
                                                n_spinorbitals=self.molecule.n_active_sos,
                                                n_electrons=self.molecule.n_active_electrons,
                                                up_then_down=self.up_then_down,
                                                spin=self.molecule.spin)

            self.core_constant, self.oneint, self.twoint = self.molecule.get_active_space_integrals()

            self.qubit_hamiltonian = qubitop_to_qubitham(qubit_op, self.qubit_mapping, self.up_then_down)

            if self.penalty_terms:
                pen_ferm = combined_penalty(self.molecule.n_active_mos, self.penalty_terms)
                pen_qubit = fermion_to_qubit_mapping(fermion_operator=pen_ferm,
                                                     mapping=self.qubit_mapping,
                                                     n_spinorbitals=self.molecule.n_active_sos,
                                                     n_electrons=self.molecule.n_active_electrons,
                                                     up_then_down=self.up_then_down,
                                                     spin=self.molecule.spin)
                pen_qubit = qubitop_to_qubitham(pen_qubit, self.qubit_hamiltonian.mapping, self.qubit_hamiltonian.up_then_down)
                self.qubit_hamiltonian += pen_qubit

            # Build / set ansatz circuit. Use user-provided circuit or built-in ansatz depending on user input.
            if isinstance(self.ansatz, BuiltInAnsatze):
                if self.ansatz == BuiltInAnsatze.UCCSD:
                    self.ansatz = UCCSD(self.molecule, self.qubit_mapping, self.up_then_down, self.molecule.spin)
                elif self.ansatz == BuiltInAnsatze.HEA:
                    self.ansatz_options["reference_state"] = "zero"
                    self.ansatz = HEA(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                elif self.ansatz == BuiltInAnsatze.UpCCGSD:
                    self.ansatz = UpCCGSD(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
                elif self.ansatz == BuiltInAnsatze.UCCGD:
                    self.ansatz = UCCGD(self.molecule, self.qubit_mapping, up_then_down=self.up_then_down)
                else:
                    raise ValueError(f"Unsupported ansatz for SA_VQESolver. Built-in ansatze:\n\t{self.builtin_ansatze}")
            elif not isinstance(self.ansatz, Ansatz):
                raise TypeError(f"Invalid ansatz dataype. Expecting instance of Ansatz class, or one of built-in options:\n\t{self.builtin_ansatze}")

        # Building with a qubit Hamiltonian.
        elif self.ansatz == BuiltInAnsatze.HEA:
            self.ansatz = HEA(self.molecule, self.qubit_mapping, self.up_then_down, **self.ansatz_options)
        elif not isinstance(self.ansatz, Ansatz):
            raise TypeError(f"Invalid ansatz dataype. Expecting a custom Ansatz (Ansatz class).")

        self.ansatz.default_reference_state = "zero"
        self.reference_circuits = list()
        for ref_state in self.ref_states:
            if isinstance(ref_state, Circuit):
                self.reference_circuits.append(ref_state)
            else:
                mapped_state = statevector_mapping.get_mapped_vector(ref_state, self.qubit_mapping, self.up_then_down)
                self.reference_circuits.append(statevector_mapping.vector_to_circuit(mapped_state))
                self.reference_circuits[-1].name = str(ref_state)

        # Set ansatz initial parameters (default or use input), build corresponding ansatz circuit
        self.initial_var_params = self.ansatz.set_var_params(self.initial_var_params)
        self.ansatz.build_circuit()

        # Quantum circuit simulation backend options
        t = self.backend_options.get("target", self.default_backend_options["target"])
        ns = self.backend_options.get("n_shots", self.default_backend_options["n_shots"])
        nm = self.backend_options.get("noise_model", self.default_backend_options["noise_model"])
        self.backend = Simulator(target=t, n_shots=ns, noise_model=nm)

    def simulate(self):
        """Run the SA-VQE algorithm, using the ansatz, classical optimizer, initial
        parameters and hardware backend built in the build method.
        """
        if not (self.ansatz and self.backend):
            raise RuntimeError(f"No ansatz circuit or hardware backend built. Have you called {self.__class__.build} ?")
        optimal_energy, optimal_var_params = self.optimizer(self.energy_estimation, self.initial_var_params)

        self.optimal_var_params = optimal_var_params
        self.optimal_energy = optimal_energy
        self.ansatz.build_circuit(self.optimal_var_params)
        self.optimal_circuit = self.ansatz.circuit
        return self.optimal_energy

    def energy_estimation(self, var_params):
        """Estimate state-averaged energy using the given ansatz, qubit hamiltonian and compute
        backend. Keeps track of optimal energy and variational parameters along
        the way.

        Args:
             var_params (numpy.array or str): variational parameters to use for
                VQE energy evaluation.

        Returns:
             float: energy computed by VQE using the ansatz and input
                variational parameters.
        """

        # Update variational parameters, compute energy using the hardware backend
        self.ansatz.update_var_params(var_params)
        energy = 0
        self.state_energies = list()
        for i, reference_circuit in enumerate(self.reference_circuits):
            state_energy = self.backend.get_expectation_value(self.qubit_hamiltonian, reference_circuit + self.ansatz.circuit)
            for circ in self.deflation_circuits:
                f_dict, _ = self.backend.simulate(circ + self.ansatz.circuit.inverse() + reference_circuit.inverse())
                state_energy += self.deflation_coeff * f_dict.get("0"*self.ansatz.circuit.width, 0)
            energy += state_energy*self.weights[i]
            self.state_energies.append(state_energy)

        if self.verbose:
            print(f"\tEnergy = {energy:.7f} ")

        return energy



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines various kinds of operators used in vqe. It can later be
broken down in several modules if needed.
"""

from math import sqrt
from collections import OrderedDict

# Later on, if needed, we can extract the code for the operators themselves to remove the dependencies and customize
import openfermion


class FermionOperator(openfermion.FermionOperator):
    """Currently, this class is coming from openfermion. Can be later on be
    replaced by our own implementation.
    """
    pass


class QubitOperator(openfermion.QubitOperator):
    """Currently, this class is coming from openfermion. Can be later on be
    replaced by our own implementation.
    """

    def frobenius_norm_compression(self, epsilon, n_qubits):
        """Reduces the number of operator terms based on its Frobenius norm
        and a user-defined threshold, epsilon. The eigenspectrum of the
        compressed operator will not deviate more than epsilon. For more
        details, see J. Chem. Theory Comput. 2020, 16, 2, 1055–1063.

        Args:
            epsilon (float): Parameter controlling the degree of compression
                and resulting accuracy.
            n_qubits (int): Number of qubits in the register.

        Returns:
            QubitOperator: The compressed qubit operator.
        """

        compressed_op = dict()
        coef2_sum = 0.
        frob_factor = 2**(n_qubits // 2)

        # Arrange the terms of the qubit operator in ascending order
        self.terms = OrderedDict(sorted(self.terms.items(), key=lambda x: abs(x[1]), reverse=False))

        for term, coef in self.terms.items():
            coef2_sum += abs(coef)**2
            # while the sum is less than epsilon / factor, discard the terms
            if sqrt(coef2_sum) > epsilon / frob_factor:
                compressed_op[term] = coef
        self.terms = compressed_op
        self.compress()


class QubitHamiltonian(QubitOperator):
    """QubitHamiltonian objects are essentially openfermion.QubitOperator
    objects, with extra attributes. The mapping procedure (mapping) and the
    qubit ordering (up_then_down) are incorporated into the class. In addition
    to QubitOperator, several checks are done when performing arithmetic
    operations on QubitHamiltonians.

    Attributes:
        term (openfermion-like): Same as openfermion term formats.
        coefficient (complex): Coefficient for this term.
        mapping (string): Mapping procedure for fermionic to qubit encoding
            (ex: "JW", "BK", etc.).
        up_then_down (bool): Whether or not spin ordering is all up then
            all down.

    Properties:
        n_terms (int): Number of terms in this qubit Hamiltonian.
    """

    def __init__(self, term=None, coefficient=1., mapping=None, up_then_down=None):
        super(QubitOperator, self).__init__(term, coefficient)

        self.mapping = mapping
        self.up_then_down = up_then_down

    @property
    def n_terms(self):
        return len(self.terms)

    def __iadd__(self, other_hamiltonian):

        # Raise error if attributes are not the same across Hamiltonians. This
        # check is ignored if comparing to a QubitOperator or a bare
        # QubitHamiltonian.
        if self.mapping is not None and self.up_then_down is not None and \
                                other_hamiltonian.mapping is not None and \
                                other_hamiltonian.up_then_down is not None:

            if self.mapping.upper() != other_hamiltonian.mapping.upper():
                raise RuntimeError("Mapping must be the same for all QubitHamiltonians.")
            elif self.up_then_down != other_hamiltonian.up_then_down:
                raise RuntimeError("Spin ordering must be the same for all QubitHamiltonians.")

        return super(QubitOperator, self).__iadd__(other_hamiltonian)

    def __eq__(self, other_hamiltonian):

        # Additional checks for == operator. This check is ignored if comparing
        # to a QubitOperator or a bare QubitHamiltonian.
        if self.mapping is not None and self.up_then_down is not None and \
                                other_hamiltonian.mapping is not None and \
                                other_hamiltonian.up_then_down is not None:
            if (self.mapping.upper() != other_hamiltonian.mapping.upper()) or (self.up_then_down != other_hamiltonian.up_then_down):
                return False

        return super(QubitOperator, self).__eq__(other_hamiltonian)

    def to_qubitoperator(self):
        qubit_op = QubitOperator()
        qubit_op.terms = self.terms.copy()

        return qubit_op


def count_qubits(qb_op):
    """Return the number of qubits used by the qubit operator based on the
    highest index found in the terms.
    """
    if (len(qb_op.terms.keys()) == 0) or ((len(qb_op.terms.keys()) == 1) and (len(list(qb_op.terms.keys())[0]) == 0)):
        return 0
    else:
        return max([(sorted(pw))[-1][0] for pw in qb_op.terms.keys() if len(pw) > 0]) + 1


def normal_ordered(fe_op):
    """ Input: a Fermionic operator of class
    toolboxes.operators.FermionicOperator or openfermion.FermionicOperator for
    reordering.

    Returns:
        FermionicOperator: Normal ordered operator.
    """

    # Obtain normal ordered fermionic operator as list of terms
    norm_ord_terms = openfermion.transforms.normal_ordered(fe_op).terms

    # Regeneratore full operator using class of tangelo.toolboxes.operators.FermionicOperator
    norm_ord_fe_op = FermionOperator()
    for term in norm_ord_terms:
        norm_ord_fe_op += FermionOperator(term, norm_ord_terms[term])
    return norm_ord_fe_op


def squared_normal_ordered(all_terms):
    """Input: a list of terms to generate toolboxes.operators.FermionOperator
    or openfermion.FermionOperator

    Returns:
        FermionOperator: squared (i.e. fe_op*fe_op) and
            normal ordered.
    """

    # Obtain normal ordered fermionic operator as list of terms
    fe_op = list_to_fermionoperator(all_terms)
    fe_op *= fe_op
    return normal_ordered(fe_op)


def list_to_fermionoperator(all_terms):
    """Input: a list of terms to generate FermionOperator

    Returns:
        FermionOperator: Single merged operator.
    """

    fe_op = FermionOperator()
    for item in all_terms:
        fe_op += FermionOperator(item[0], item[1])
    return fe_op


def qubitop_to_qubitham(qubit_op, mapping, up_then_down):
    """Function to convert a QubitOperator into a QubitHamiltonian.

    Args:
        qubit_op (QubitOperator): Self-explanatory.
        mapping (string): Qubit mapping procedure.
        up_then_down (bool): Whether or not spin ordering is all up then
            all down.

    Returns:
        QubitHamiltonian: Self-explanatory.
    """
    qubit_ham = QubitHamiltonian(mapping=mapping, up_then_down=up_then_down)
    qubit_ham.terms = qubit_op.terms.copy()

    return qubit_ham



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file provides an API enabling the use of derandomized classical shadows.
This algorithm is described in H.-Y. Huang, R. Kueng, and J. Preskill,
ArXiv:2103.07510 [Quant-Ph] (2021).
"""

from math import floor, exp, log
import random

from tangelo.toolboxes.measurements import ClassicalShadow
from tangelo.linq.circuit import Circuit
from tangelo.linq.helpers.circuits.measurement_basis import measurement_basis_gates, pauli_string_to_of


class DerandomizedClassicalShadow(ClassicalShadow):
    """Classical shadows using derandomized single Pauli measurements, as
    defined in H.-Y. Huang, R. Kueng, and J. Preskill, ArXiv:2103.07510
    [Quant-Ph] (2021).
    """

    def build(self, n_shots, qu_op, eta=0.9):
        """Derandomized sampling of single Pauli words.

        Args:
            n_shots (int): Total number of measurements.
            qu_op (QubitOperator): Relevant QubitOperator.
            eta (float): Empirical parameter for the cost function. Default is
                0.9.

        Returns:
            list of str: Measurements generated for a derandomized procedure.
        """

        # Getting the weights, proportional to the coefficient of each Pauli
        # words. Some variables are defined to normalize the weights and track
        # the amount of measurements already defined by the algorithm.
        observables, weights = zip(*[(obs, abs(w)) for obs, w in qu_op.terms.items() if obs])

        n_observables = len(observables)
        n_measurements_per_observable = floor(n_shots / n_observables)

        norm_factor = n_observables / sum(weights)
        weights = [w * norm_factor for w in weights]

        weighted_n_measurements_per_observable = [floor(weights[i]*n_measurements_per_observable) for i in range(n_observables)]

        n_measurements_so_far = [0] * n_observables

        # Output variable (containing all chosen basis).
        measurement_procedure = list()

        # Distribution of n_measurements_per_observable * n_observables shots.
        for _ in range(n_measurements_per_observable * n_observables):

            # A single round of parallel measurements over all qubits.
            n_matches_needed_round = [len(p) for p in observables]

            single_round_measurement = [None] * self.n_qubits

            # Optimizing which Pauli basis to use for each qubit according to
            # self._cost_function.
            for i_qubit in range(self.n_qubits):
                cost_of_outcomes = {"X": 0, "Y": 0, "Z": 0}

                # Computing the cost function with all the possibilities.
                for dice_roll_pauli in ["Z", "X", "Y"]:
                    # Assume the dice rollout to be dice_roll_pauli.
                    try_matches_needed_round = n_matches_needed_round.copy()
                    for i_obs, obs in enumerate(observables):
                        try_matches_needed_round[i_obs] += _get_match_up(i_qubit, dice_roll_pauli, obs, self.n_qubits)

                    cost_of_outcomes[dice_roll_pauli] = self._cost_function(n_measurements_so_far,
                                                                            try_matches_needed_round,
                                                                            weights,
                                                                            weighted_n_measurements_per_observable,
                                                                            eta)

                # Determining the single Pauli gate to use.
                for dice_roll_pauli in ["Z", "X", "Y"]:
                    if min(cost_of_outcomes.values()) < cost_of_outcomes[dice_roll_pauli]:
                        continue

                    # The best dice roll outcome will be chosen here.
                    single_round_measurement[i_qubit] = dice_roll_pauli

                    for i_obs, obs in enumerate(observables):
                        n_matches_needed_round[i_obs] += _get_match_up(i_qubit, dice_roll_pauli, obs, self.n_qubits)
                    break

            measurement_procedure.append(single_round_measurement)

            # Incrementing the number of measurements so far if there is no more
            # matches to make this round.
            n_measurements_so_far = [n[0] + 1 if n[1] == 0 else n[0] for n in zip(n_measurements_so_far, n_matches_needed_round)]

            # Incrementing success variable if number of measurements so far is
            # bigger than the weighted number of measurements per observable.
            success = sum([1 for i in range(n_observables) if n_measurements_so_far[i] >= weighted_n_measurements_per_observable[i]])

            if success == n_observables:
                break

        measurement_procedure = ["".join(m) for m in measurement_procedure]

        # Fill "missing" shots with a set of random (already chosen) basis.
        measurement_procedure += random.choices(measurement_procedure, k=n_shots-len(measurement_procedure))

        self.unitaries += measurement_procedure
        return measurement_procedure

    def get_basis_circuits(self, only_unique=False):
        """Outputs a list of circuits corresponding to the chosen single-Pauli
        unitaries.

        Args:
            only_unique (bool): Considering only unique unitaries.

        Returns:
            list of Circuit or tuple: All basis circuits or a tuple of unique
                circuits, string representation and numbers of occurence.
        """

        if not self.unitaries:
            raise ValueError(f"A set of unitaries must de defined (can be done with the build method in {self.__class__.__name__}).")

        unitaries_to_convert = self.unique_unitaries if only_unique else self.unitaries

        basis_circuits = list()
        for pauli_word in unitaries_to_convert:
            # Transformation of a unitary to quantum gates.
            pauli_of = pauli_string_to_of(pauli_word)
            basis_circuits += [Circuit(measurement_basis_gates(pauli_of), self.n_qubits)]

        # Counting each unique circuits (use for reversing to a full shadow from an experiement on hardware).
        if only_unique:
            unique_basis_circuits = [(basis_circuits[i], u, self.unitaries.count(u)) for i, u in enumerate(unitaries_to_convert)]
            return unique_basis_circuits
        else:
            return basis_circuits

    def get_term_observable(self, term, coeff=1.):
        """Returns the estimated observable for a term and its coefficient.

        Args:
            term (tuple): Openfermion style of a qubit operator term.
            coeff (float): Multiplication factor for the term.

        Returns:
            float: Observable estimated with the shadow.
        """

        sum_product = 0
        n_match = 0

        # For every single_measurement in shadow_size.
        for snapshot in range(self.size):
            match = 1
            product = 1

            # Checks if there is a match for all Pauli gate in the term. Works
            # also with operator not on all qubits (e.g. X1 will hit Z0X1, Y0X1
            # and Z0X1).
            for i_qubit, pauli in term:
                if pauli != self.unitaries[snapshot][i_qubit]:
                    match = 0
                    break
                if self.bitstrings[snapshot][i_qubit] != "0":
                    product *= -1

            # No quantity is considered if there is no match.
            sum_product += match * product
            n_match += match

        return sum_product / n_match * coeff if n_match > 0 else 0.

    def _cost_function(self, n_measurements_so_far, n_matches_needed, weights, weighted_num_measurements_per_observable, eta=0.9):
        """Cost function for derandomized Pauli measurements, according to
        equation (6) in the cited paper.

        Args:
            n_measurements_so_far (list of int): Number of measurements decided
                per terms.
            n_matches_needed (list of int): Number of matches.
            weights (list of float): Coefficient (absolute) of each each term.
            weighted_num_measurements_per_observable (list of float): Weighted
                number of measurements per term.
            eta (float): Empirical parameter, default set to 0.9.

        Returns:
            float: Output the cost considering the arguments.
        """

        # Computing the cost. Variables names are consistent of what is found in
        # the publication.
        nu = 1 - exp(-eta / 2)

        cost = 0.
        for (n_measure, n_match, weight, weight_per_obs) in zip(n_measurements_so_far,
                                                                n_matches_needed,
                                                                weights,
                                                                weighted_num_measurements_per_observable):

            if n_measure >= weight_per_obs:
                continue

            v = eta / 2 * n_measure
            if self.n_qubits >= n_match:
                v -= log(1 - nu / (3**n_match))

            cost += exp(-v / weight)

        return cost


def _get_match_up(lookup_qubit, dice_roll_pauli, observable, n_qubits):
    """Helper function to output 0, -1 or a large number depending on the
    index provided and the Pauli gate in a single observable.

    Args:
        lookup_qubit (int): Qubit index.
        dice_roll_pauli (str): Z, X or Y.
        observable (tuple): Single term in the form ((0, "Y"), (1, "Z"),
            (2, "X"), ...).
        n_qubits (int): Number of qubits.

    Returns:
        int: 0, -1, or a large int used in computing cost.
    """

    large_number = 100 * (n_qubits + 10)

    for i_qubit, pauli in observable:
        if lookup_qubit == i_qubit:
            return -1 if pauli == dice_roll_pauli else large_number
    return 0



# Copyright 2021 Good Chemistry Company.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from tangelo.problem_decomposition.oniom.oniom_problem_decomposition import ONIOMProblemDecomposition
from tangelo.problem_decomposition.oniom._helpers.helper_classes import Fragment
from tangelo.molecule_library import xyz_H4
from tangelo.algorithms.variational import BuiltInAnsatze


class ONIOMQuantumTest(unittest.TestCase):

    def test_energy_hf_vqe_uccsd_h4(self):
        """Test to verifiy the implementation of VQE (with UCCSD) in ONIOM."""

        options_hf = {"basis": "sto-3g"}
        options_vqe = {"basis": "sto-3g", "ansatz": BuiltInAnsatze.UCCSD}

        # With this line, the interaction between H2-H2 is computed with a low
        # accuracy method.
        system = Fragment(solver_low="HF", options_low=options_hf)
        # VQE-UCCSD fragments.
        model_vqe_1 = Fragment(solver_low="HF",
                               options_low=options_hf,
                               solver_high="VQE",
                               options_high=options_vqe,
                               selected_atoms=[0, 1])
        model_vqe_2 = Fragment(solver_low="HF",
                               options_low=options_hf,
                               solver_high="VQE",
                               options_high=options_vqe,
                               selected_atoms=[2, 3])
        oniom_model_vqe = ONIOMProblemDecomposition({"geometry": xyz_H4, "fragments": [system, model_vqe_1, model_vqe_2]})

        e_oniom_vqe = oniom_model_vqe.simulate()

        # ONIOM + CCSD is tested in test_oniom.ONIOMTest.test_energy_hf_ccsd_h4.
        self.assertAlmostEqual(-1.901616, e_oniom_vqe, places=5)

    def test_energy_hf_qite_h4(self):
        """Test to verifiy the implementation of QITE in ONIOM."""

        options_both = {"basis": "sto-3g", "verbose": False}

        # With this line, the interaction between H2-H2 is computed with a low
        # accuracy method.
        system = Fragment(solver_low="HF", options_low=options_both)
        # QITE fragments.
        model_qite_1 = Fragment(solver_low="HF",
                               options_low=options_both,
                               solver_high="QITE",
                               options_high=options_both,
                               selected_atoms=[0, 1])
        model_qite_2 = Fragment(solver_low="HF",
                               options_low=options_both,
                               solver_high="QITE",
                               options_high=options_both,
                               selected_atoms=[2, 3])
        oniom_model_qite = ONIOMProblemDecomposition({"geometry": xyz_H4, "fragments": [system, model_qite_1, model_qite_2]})

        e_oniom_qite = oniom_model_qite.simulate()

        # ONIOM + CCSD is tested in test_oniom.ONIOMTest.test_energy_hf_ccsd_h4.
        self.assertAlmostEqual(-1.901616, e_oniom_qite, places=5)

    def test_energy_hf_adapt_h4(self):
        """Test to verifiy the implementation of ADAPT-VQE in ONIOM."""

        options_both = {"basis": "sto-3g", "verbose": False}

        # With this line, the interaction between H2-H2 is computed with a low
        # accuracy method.
        system = Fragment(solver_low="HF", options_low=options_both)
        # ADAPT fragments.
        model_adapt_1 = Fragment(solver_low="HF",
                               options_low=options_both,
                               solver_high="ADAPT",
                               options_high=options_both,
                               selected_atoms=[0, 1])
        model_adapt_2 = Fragment(solver_low="HF",
                               options_low=options_both,
                               solver_high="ADAPT",
                               options_high=options_both,
                               selected_atoms=[2, 3])
        oniom_model_adapt = ONIOMProblemDecomposition({"geometry": xyz_H4, "fragments": [system, model_adapt_1, model_adapt_2]})

        e_oniom_adapt = oniom_model_adapt.simulate()

        # ONIOM + CCSD is tested in test_oniom.ONIOMTest.test_energy_hf_ccsd_h4.
        self.assertAlmostEqual(-1.901616, e_oniom_adapt, places=5)

    def test_get_resources_vqe(self):
        """Test to verifiy the implementation of resources estimation (VQE) in
        ONIOM. Other quantum solvers should also work if VQE works.
        """

        options_hf = {"basis": "sto-3g"}
        options_vqe = {"basis": "sto-3g", "ansatz": BuiltInAnsatze.UCCSD, "qubit_mapping": "JW"}

        system = Fragment(solver_low="HF", options_low=options_hf)
        # VQE-UCCSD fragments.
        model_vqe_1 = Fragment(solver_low="HF",
                               options_low=options_hf,
                               solver_high="VQE",
                               options_high=options_vqe,
                               selected_atoms=[0, 1])
        model_vqe_2 = Fragment(solver_low="HF",
                               options_low=options_hf,
                               solver_high="VQE",
                               options_high=options_vqe,
                               selected_atoms=[2, 3])
        oniom_model_vqe = ONIOMProblemDecomposition({"geometry": xyz_H4, "fragments": [system, model_vqe_1, model_vqe_2]})

        res = oniom_model_vqe.get_resources()

        self.assertEqual(res[1]["circuit_width"], 4)
        self.assertEqual(res[1]["qubit_hamiltonian_terms"], 15)

        self.assertEqual(res[2]["circuit_width"], 4)
        self.assertEqual(res[2]["qubit_hamiltonian_terms"], 15)


if __name__ == "__main__":
    unittest.main()
